# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

"""Test domain corner cases."""
import unittest

from mailmanclient import Client
from six.moves.urllib_error import HTTPError


__metaclass__ = type
__all__ = [
    'TestDomains',
    ]


class TestDomains(unittest.TestCase):
    def setUp(self):
        self._client = Client(
            'http://localhost:9001/3.0', 'restadmin', 'restpass')

    def test_no_domain(self):
        # Trying to get a non-existent domain returns a 404.
        #
        # We can't use `with self.assertRaises()` until we drop Python 2.6
        try:
            self._client.get_domain('example.org')
        except HTTPError as error:
            self.assertEqual(error.code, 404)
        else:
            raise AssertionError('Expected HTTPError 404')



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from mailmanclient.restbase.base import RESTDict

__metaclass__ = type
__all__ = [
    'ListArchivers'
]


class ListArchivers(RESTDict):
    """
    Represents the activation status for each site-wide available archiver
    for a given list.
    """

    _autosave = True

    def __init__(self, connection, url, mlist):
        """
        :param connection: An API connection object.
        :type connection: Connection.
        :param url: The API url of the list's archiver endpoint.
        :type url: str.
        :param mlist: The corresponding list object.
        :type mlist: MailingList.
        """
        super(ListArchivers, self).__init__(connection, url)
        self._mlist = mlist

    def __repr__(self):
        return '<Archivers on {0!r}>'.format(self._mlist.list_id)

    def __str__(self):
        return 'Archivers on {}'.format(self._mlist.list_id)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from mailmanclient.restbase.base import RESTDict

__metaclass__ = type
__all__ = [
    'Preferences',
    'PreferencesMixin'
]


class Preferences(RESTDict):

    _properties = (
        'acknowledge_posts', 'delivery_mode', 'delivery_status',
        'hide_address', 'preferred_language', 'receive_list_copy',
        'receive_own_postings',
        )

    def delete(self):
        response, content = self._connection.call(self._url, method='DELETE')


class PreferencesMixin:
    """Mixin for restobjects that have preferences."""

    @property
    def preferences(self):
        if getattr(self, '_preferences', None) is None:
            path = '{0}/preferences'.format(self.self_link)
            self._preferences = Preferences(self._connection, path)
        return self._preferences



# Copyright (C) 2019-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Test for the general Client specific behavior.."""

import unittest

from mailmanclient import Client
from urllib.error import HTTPError


class TestUrlencodedPaths(unittest.TestCase):
    """Test that paths are always safe"""

    def setUp(self):
        self._client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass')
        self.domain = self._client.create_domain('example.org')

    def tearDown(self):
        self.domain.delete()
        for user in self._client.users:
            user.delete()

    def test_bans_paths_are_urlencoded_when_needed(self):
        # Test that we can ban regular expressions.
        self._client.bans.add('^something?example@something.com')
        bans = self._client.bans
        self.assertEqual(len(bans), 1)
        # This should actually call the URL and should return True.
        self.assertIn('^something?example@something.com', self._client.bans)

    def test_member_paths_are_urlencoded(self):
        mlist = self.domain.create_list('other')
        mlist.subscribe('apers?on@example.com',
                        pre_verified=True,
                        pre_confirmed=True,
                        pre_approved=True)
        # Make sure that member exists.
        self.assertEqual(len(mlist.members), 1)
        # Let's see if we can get the member resource. This will
        try:
            member = mlist.get_member('apers?on@example.com')
            self.assertIsNotNone(member)
        except HTTPError:
            self.fail('Unexpected HTTPError.')

    def test_non_member_paths_are_urlencoded(self):
        mlist = self.domain.create_list('some')
        mlist.add_role(address='somperson?123@example.com', role='moderator')
        # Now, this shouldn't raise 404 error.
        try:
            mlist.remove_role(
                address='somperson?123@example.com', role='moderator')
        except HTTPError:
            self.fail('Unexpected HTTPError.')


class TestHooks(unittest.TestCase):

    def test_initialize_with_hooks(self):
        # Test that we can initialize mailmanclient with hooks.
        calls = []

        def store_calls(params):
            calls.append(params)
            return params
        client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass',
            request_hooks=[store_calls])
        # Make *any* API call.
        client.domains
        # Now make sure the call was stored.
        self.assertEqual(len(calls), 1)
        params = calls[0]
        self.assertEqual(
            params.get('url'), 'http://localhost:9001/3.1/domains')

    def test_add_hooks(self):
        calls = []

        def store_calls(params):
            calls.append(params)
            return params
        client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass')
        # Make *any* API call.
        client.domains
        # Now make sure the call was not stored.
        self.assertEqual(len(calls), 0)
        client.add_hooks([store_calls])
        # Now make another API call.
        client.lists
        params = calls[0]
        self.assertEqual(params.get('url'), 'http://localhost:9001/3.1/lists')


class TestFindUsers(unittest.TestCase):

    def setUp(self):
        self._client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass')
        self._client.create_user('anne@example.com', 'xxx', 'Anne Person')
        self._client.create_user('bart@example.com', 'xxx', 'Bee Person')

    def tearDown(self):
        for user in self._client.users:
            user.delete()

    def test_find_users(self):
        # Make sure the users are created.
        self.assertEqual(len(self._client.users), 2)
        # Now search for the user.
        users = self._client.find_users('bee')
        self.assertEqual(len(users), 1)
        self.assertEqual(users[0].display_name, 'Bee Person')
        # search for the email.
        users = self._client.find_users('ne@ex')
        self.assertEqual(len(users), 1)
        self.assertEqual(users[0].display_name, 'Anne Person')

    def test_find_user_page(self):
        self.assertEqual(len(self._client.users), 2)
        # Now search for the user.
        users = self._client.find_users_page('person', count=1, page=1)
        # We should've got only single user with count=1.
        self.assertEqual(len(users), 1)
        # Total results should be 2.
        self.assertEqual(users.total_size, 2)
        self.assertEqual(users[0].display_name, 'Anne Person')



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

from urllib.parse import quote_plus

from mailmanclient.restobjects.preferences import PreferencesMixin
from mailmanclient.restbase.base import RESTList, RESTObject

__metaclass__ = type
__all__ = [
    'Address',
    'Addresses'
]


class Addresses(RESTList):

    def __init__(self, connection, url, data=None):
        super(Addresses, self).__init__(connection, url, data)
        self._factory = lambda data: Address(
            self._connection, data['self_link'], data)

    def find_by_email(self, email):
        for address in self:
            if address.email == email:
                return address
        return None

    def remove(self, email):
        address = self.find_by_email(email)
        if address is not None:
            address.delete()
            self._reset_cache()
        else:
            raise ValueError('The address {} does not exist'.format(email))


class Address(RESTObject, PreferencesMixin):

    _properties = ('display_name', 'email', 'original_email', 'registered_on',
                   'self_link', 'verified_on')

    def __repr__(self):
        return '<Address {!r}>'.format(self.email)

    def __str__(self):
        return self.email

    @property
    def user(self):
        from mailmanclient.restobjects.user import User
        if 'user' in self.rest_data:
            return User(self._connection, self.rest_data['user'])
        else:
            return None

    @property
    def verified(self):
        return self.verified_on is not None

    def verify(self):
        self._connection.call(
            'addresses/{0}/verify'.format(quote_plus(self.email)),
            method='POST',
            )
        self._reset_cache()

    def unverify(self):
        self._connection.call(
            'addresses/{0}/unverify'.format(quote_plus(self.email)),
            method='POST'
            )
        self._reset_cache()



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""setup.py helper functions."""

from __future__ import absolute_import, print_function, unicode_literals

import os
import re
import sys
import codecs


__metaclass__ = type
__all__ = [
    'description',
    'find_doctests',
    'get_version',
    'long_description',
    'require_python',
    ]


DEFAULT_VERSION_RE = re.compile(r'(?P<version>\d+\.\d(?:\.\w+)?)')
NL = '\n'


def require_python(minimum):
    """Require at least a minimum Python version.

    The version number is expressed in terms of `sys.hexversion`.  E.g. to
    require a minimum of Python 2.6, use::

    >>> require_python(0x206000f0)

    :param minimum: Minimum Python version supported.
    :type minimum: integer
    """
    if sys.hexversion < minimum:
        hversion = hex(minimum)[2:]
        if len(hversion) % 2 != 0:
            hversion = '0' + hversion
        split = list(hversion)
        parts = []
        while split:
            parts.append(int(''.join((split.pop(0), split.pop(0))), 16))
        major, minor, micro, release = parts
        if release == 0xf0:
            print('Python {0}.{1}.{2} or better is required'.format(
                major, minor, micro))
        else:
            print('Python {0}.{1}.{2} ({3}) or better is required'.format(
                major, minor, micro, hex(release)[2:]))
        sys.exit(1)


def get_version(filename, pattern=None):
    """Extract the __version__ from a file without importing it.

    While you could get the __version__ by importing the module, the very act
    of importing can cause unintended consequences.  For example, Distribute's
    automatic 2to3 support will break.  Instead, this searches the file for a
    line that starts with __version__, and extract the version number by
    regular expression matching.

    By default, two or three dot-separated digits are recognized, but by
    passing a pattern parameter, you can recognize just about anything.  Use
    the `version` group name to specify the match group.

    :param filename: The name of the file to search.
    :type filename: string
    :param pattern: Optional alternative regular expression pattern to use.
    :type pattern: string
    :return: The version that was extracted.
    :rtype: string
    """
    if pattern is None:
        cre = DEFAULT_VERSION_RE
    else:
        cre = re.compile(pattern)
    with open(filename) as fp:
        for line in fp:
            if line.startswith('__version__'):
                mo = cre.search(line)
                assert mo, 'No valid __version__ string found'
                return mo.group('version')
    raise AssertionError('No __version__ assignment found')


def find_doctests(start='.', extension='.txt'):
    """Find separate-file doctests in the package.

    This is useful for Distribute's automatic 2to3 conversion support.  The
    `setup()` keyword argument `convert_2to3_doctests` requires file names,
    which may be difficult to track automatically as you add new doctests.

    :param start: Directory to start searching in (default is cwd)
    :type start: string
    :param extension: Doctest file extension (default is .txt)
    :type extension: string
    :return: The doctest files found.
    :rtype: list
    """
    doctests = []
    for dirpath, _, filenames in os.walk(start):
        doctests.extend(os.path.join(dirpath, filename)
                        for filename in filenames
                        if filename.endswith(extension))
    return doctests


def long_description(*filenames):
    """Provide a long description."""
    res = []
    for value in filenames:
        base, ext = os.path.splitext(value)
        if ext in ('.txt', '.rst'):
            with codecs.open(value, 'r', encoding='utf-8') as fp:
                value = fp.read()
        res.append(value)
        if not value.endswith(NL):
            res.append('')
    return NL.join(res)


def description(filename):
    """Provide a short description."""
    with codecs.open(filename, 'r', encoding='utf-8') as fp:
        for line in fp:
            return line.strip()



# coding: utf-8

# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.


"""Test unicode data when using REST endpoint."""
import unittest

from mailmanclient import Client
from six.moves.urllib_error import HTTPError


__metaclass__ = type
__all__ = [
    'TestUnicode',
    ]


class TestUnicode(unittest.TestCase):
    def setUp(self):
        self._client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass')
        self.email = 'jeremy@example.com'
        self.unicode_string = u'Jérôme'

    def tearDown(self):
        try:
            self._client.get_user(self.email).delete()
        except HTTPError as error:
            if error.status_code == 404:
                pass

    def test_create_user(self):
        user = self._client.create_user(
            email=self.email, password='1234',
            display_name=self.unicode_string)
        self.assertEqual(user.display_name, self.unicode_string)

    def test_repr(self):
        user = self._client.create_user(
            email=self.email, password='1234',
            display_name=self.unicode_string)
        try:
            repr(user)
        except UnicodeEncodeError as e:
            self.fail(e)



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Async connection object."""

__all__ = [
    'Connection',
]

from urllib.error import HTTPError
from mailmanclient.restbase.connection import Connection as BaseConnection


class Connection(BaseConnection):
    """A standard Connection object.

    This is an abstraction over HTTP connections for Mailmanclient. It can be
    initialized with any http client library with and async `request`
    method. The paramters are currently tailored for httpx, but if there are
    folks interested in others, it is easy to provide a wrapper which accept
    such parameters.

    :param client: The http client object with ``request`` method.
    """

    def __init__(self, client, *args, **kw) -> None:
        self.client = client
        super().__init__(*args, **kw)

    async def call(self, path, data=None, method=None):
        params = self._prepare_request(
            path, data, method
            )
        response = await self.client.request(auth=self.auth, **params)
        if response.status_code // 100 != 2:
            raise HTTPError(params.get('url'), response.status_code,
                            response.content, None, None)
        if len(response.content) == 0:
            return response, None
        return response, response.json()



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Preferences object and mixins."""

__all__ = [
    'Preferences',
    'PreferencesMixin',
]

from mailmanclient.restbase.async_base import RESTObject


class Preferences(RESTObject):

    _properties = (
        'acknowledge_posts', 'delivery_mode', 'delivery_status',
        'hide_address', 'preferred_language', 'receive_list_copy',
        'receive_own_postings',
        )


class PreferencesMixin:
    """Mixin for restobjects that have preferences."""

    async def preferences(self):
        if getattr(self, '_preferences', None) is None:
            path = '{0}/preferences'.format(self.self_link)
            response, content = await self._connection.call(path)
            self._preferences = Preferences(self._connection, content)
        return self._preferences



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

"""Test Page corner cases."""
import unittest

from unittest.mock import Mock
from urllib.parse import urlsplit, parse_qs

from mailmanclient.constants import DEFAULT_PAGE_ITEM_COUNT
from mailmanclient.restbase.page import Page

__metaclass__ = type
__all__ = [
    'TestPage',
    ]


class TestPage(unittest.TestCase):

    def test_url_simple(self):
        connection = Mock()
        connection.call.return_value = (None, {'start': 0, 'total_size': 0})
        page = Page(connection, '/some-path', None)
        built_qs = parse_qs(urlsplit(page._build_url()).query)
        self.assertEqual(built_qs, dict(
            count=[str(DEFAULT_PAGE_ITEM_COUNT)],
            page=["1"]))

    def test_url_with_qs(self):
        connection = Mock()
        connection.call.return_value = (None, {'start': 0, 'total_size': 0})
        page = Page(connection, '/some-path?with=a&query=string', None)
        built_qs = parse_qs(urlsplit(page._build_url()).query)
        self.assertEqual(built_qs, {
            "with": ["a"],
            "query": ["string"],
            "count": [str(DEFAULT_PAGE_ITEM_COUNT)],
            "page": ["1"],
            })



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for Mailing List."""

import time

from urllib.error import HTTPError
from unittest import TestCase
from mailmanclient import Client
from mailmanclient.restobjects.member import Member


class TestMailingListMembershipTests(TestCase):

    def setUp(self):
        self._client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass')
        try:
            self.domain = self._client.create_domain('example.com')
        except HTTPError:
            self.domain = self._client.get_domain('example.com')
        self.mlist = self.domain.create_list('foo')

    def tearDown(self):
        self.domain.delete()

    def test_list_is_owner(self):
        # Tests MailingList.is_owner
        # First, we add an owner to the mailing list and then make sure that
        # we see it in the owners roster.
        anne_addr = 'ann@example.com'
        self.mlist.add_owner(anne_addr)
        # Check that the address
        owners_list = [owner.email for owner in self.mlist.owners]
        self.assertIn(anne_addr, owners_list)
        # Now, we make sure that we get the same result in our API.
        self.assertTrue(self.mlist.is_owner(anne_addr))
        # Make sure we get False for someone who is not a list owner.
        self.assertFalse(self.mlist.is_owner('random@example.com'))
        # Make sure that a subscriber doesn't return True for is_owner check.
        # We are doing this test because of the way is_owner test works. A
        # wrong value for `role` could result in a list member being tested
        # as owner.
        self.mlist.subscribe('bart@example.com')
        self.assertFalse(self.mlist.is_owner('bart@example.com'))
        # Now, try the same thing for Moderators.
        self.mlist.add_moderator('mod@example.com')
        self.assertFalse(self.mlist.is_owner('mod@example.com'))

    def test_list_is_moderator(self):
        # Tests MailingList.is_moderator
        # First, we add a moderator to the list.
        mod_addr = 'mod@example.com'
        self.mlist.add_moderator(mod_addr)
        mods_emails = [mod.email for mod in self.mlist.moderators]
        self.assertIn(mod_addr, mods_emails)
        self.assertFalse(self.mlist.is_owner(mod_addr))
        # Owners shouldn't return true for this API.
        owner_addr = 'owner@example.com'
        self.mlist.add_owner(owner_addr)
        self.assertFalse(self.mlist.is_moderator(owner_addr))
        # Subscribers shouldn't return true for this API.
        subscriber_addr = 'subscriber@example.com'
        self.mlist.subscribe(subscriber_addr)
        self.assertFalse(self.mlist.is_moderator(subscriber_addr))

    def test_list_is_member(self):
        # Tests MailingList.is_member
        subscriber_addr = 'subscriber@example.com'
        self.mlist.subscribe(subscriber_addr, pre_verified=True,
                             pre_confirmed=True, pre_approved=True)
        all_subscribers = [member.email for member in self.mlist.members]
        self.assertIn(subscriber_addr, all_subscribers)
        # Now make sure we get the same result through this API.
        self.assertTrue(self.mlist.is_member(subscriber_addr))
        # Make sure owners don't pass this check.
        owner_addr = 'owner@example.com'
        self.mlist.add_owner(owner_addr)
        self.assertFalse(self.mlist.is_member(owner_addr))
        # Make sure moderators don't pass this check.
        mod_addr = 'mod@example.com'
        self.mlist.add_moderator(mod_addr)
        self.assertFalse(self.mlist.is_member(mod_addr))

    def test_list_is_owner_or_mod(self):
        # Tests MailingList.is_owner_or_mod
        # Tests MailingList.is_moderator
        # First, we add a moderator to the list.
        mod_addr = 'mod@example.com'
        self.mlist.add_moderator(mod_addr)
        mods_emails = [mod.email for mod in self.mlist.moderators]
        self.assertIn(mod_addr, mods_emails)
        self.assertTrue(self.mlist.is_owner_or_mod(mod_addr))
        # Owners shouldn't return true for this API.
        owner_addr = 'owner@example.com'
        self.mlist.add_owner(owner_addr)
        owners_list = [owner.email for owner in self.mlist.owners]
        self.assertIn(owner_addr, owners_list)
        self.assertTrue(self.mlist.is_owner_or_mod(owner_addr))
        # Subscribers shouldn't return true for this API.
        subscriber_addr = 'subscriber@example.com'
        self.mlist.subscribe(subscriber_addr)
        self.assertFalse(self.mlist.is_owner_or_mod(subscriber_addr))

    def test_list_unsubscribe(self):
        # Tests MailingList.unsubscribe
        subscriber_addr = 'subscriber@example.com'
        self.mlist.subscribe(subscriber_addr, pre_verified=True,
                             pre_confirmed=True, pre_approved=True)
        self.assertTrue(self.mlist.is_member(subscriber_addr))
        self.mlist.unsubscribe(subscriber_addr, pre_confirmed=True,
                               pre_approved=True)
        self.assertFalse(self.mlist.is_member(subscriber_addr))


class TestHeldMessage(TestCase):

    def setUp(self):
        self._client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass')

        try:
            self.domain = self._client.create_domain('example.com')
        except HTTPError:
            self.domain = self._client.get_domain('example.com')

        self.mlist = self.domain.create_list('foo')
        # Test that a held message can be moderated.
        msg = """\
From: nonmember@example.com
To: foo@example.com
Subject: Hello World
Message-ID: <msgid>

Hello!
"""
        self._inject_message(msg, self.mlist)
        # Wait for the message to appear in the held queue for 10 seconds max.
        self._wait_for_message_in_held_queue(self.mlist, 30)

    def tearDown(self):
        self.domain.delete()

    def _inject_message(self, msg, mlist):
        inq = self._client.queues['in']
        inq.inject('foo.example.com', msg)

    def _wait_for_message_in_held_queue(self, mlist, timeout):
        """Wait for held message in mlist for timeout seconds."""
        start_time = time.time()
        while True:
            # if (start_time + timeout > time.time()):
            #     print('timeout trying to wait for message')
            #     break
            all_held = mlist.held
            if len(all_held) > 0:
                print('Total time to wait for message:')
                print(time.time() - start_time)
                break
            time.sleep(0.1)

    def test_held_message_moderation(self):
        # Test that message was held and not timed out.
        self.assertEqual(len(self.mlist.held), 1)
        held = self.mlist.held[0]
        held_message = self.mlist.get_held_message(held.request_id)

        # Now, let's try to reject this message with a reason.
        response = self.mlist.reject_message(held_message.request_id,
                                             reason='You shall not pass.')
        self.assertEqual(response.status_code, 204)
        # Make sure that the message was rejected.
        self.assertEqual(len(self.mlist.held), 0)

    def test_held_message_moderation_message_object(self):
        # The only difference between this and the above test is that we use
        # the `held_message.reject` API, which for some reason as different
        # code to make the same API call.
        self.assertEqual(len(self.mlist.held), 1)
        held = self.mlist.held[0]
        held_message = self.mlist.get_held_message(held.request_id)

        # Now, let's try to reject this message with a reason.
        response = held_message.reject(reason='You shall not pass.')
        self.assertEqual(response.status_code, 204)
        # Make sure that the message was rejected.
        self.assertEqual(len(self.mlist.held), 0)


class TestMailingList(TestCase):

    def setUp(self):
        self._client = Client(
            'http://localhost:9001/3.1', 'restadmin', 'restpass')
        try:
            self.domain = self._client.create_domain('example.com')
        except HTTPError:
            self.domain = self._client.get_domain('example.com')
        self.mlist = self.domain.create_list('foo')

    def tearDown(self):
        self.domain.delete()

    def test_subscribe_without_display_name(self):
        self.mlist.subscribe('aperson@example.com',
                             pre_verified=True,
                             pre_confirmed=True,
                             pre_approved=True)
        users = self.mlist.members[0]
        self.assertEqual(users.display_name, '')

    def test_subscribe_with_display_name(self):
        self.mlist.subscribe('bperson@example.com',
                             display_name='B Person',
                             pre_verified=True,
                             pre_confirmed=True,
                             pre_approved=True)
        users = self.mlist.members[0]
        self.assertEqual(users.display_name, 'B Person')

    def test_invite(self):
        data = self.mlist.subscribe('cperson@example.com',
                                    invitation=True)
        # cperson is not a member yet.
        self.assertRaisesRegex(ValueError,
                               'not a member',
                               self.mlist.get_member,
                               'cperson@example.com')
        # But we got a token for the invitation.
        self.assertEqual(data['token_owner'], 'subscriber')

    def test_get_individual_pending_request(self):
        data = self.mlist.subscribe('aperson@example.com', pre_confirmed=False)
        self.assertEqual(data['token_owner'], 'subscriber')
        # Now get the individual request object.
        json = self.mlist.get_request(data['token'])
        self.assertEqual(json['token_owner'], data['token_owner'])
        self.assertEqual(json['token'], data['token'])

    def test_get_unsubscription_requests(self):
        member = self.mlist.subscribe(
            'aperson@example.com',
            pre_confirmed=True, pre_verified=True, pre_approved=True)
        self.assertTrue(isinstance(member, Member))
        self.assertEqual(len(self.mlist.members), 1)
        settings = self.mlist.settings
        settings['unsubscription_policy'] = 'moderate'
        settings.save()
        # try to unsubscribe a user.
        self.mlist.unsubscribe('aperson@example.com', pre_approved=False)
        self.assertEqual(len(self.mlist.members), 1)
        self.assertEqual(len(self.mlist.unsubscription_requests), 1)
        unsub_req = self.mlist.unsubscription_requests[0]
        self.assertEqual(unsub_req['token_owner'], 'moderator')
        self.assertEqual(unsub_req['email'], 'aperson@example.com')



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
import warnings
from operator import itemgetter
from urllib.error import HTTPError
from urllib.parse import urlencode, quote_plus

from mailmanclient.restobjects.header_match import HeaderMatches
from mailmanclient.restobjects.archivers import ListArchivers
from mailmanclient.restobjects.member import Member
from mailmanclient.restobjects.settings import Settings
from mailmanclient.restobjects.held_message import HeldMessage
from mailmanclient.restobjects.templates import TemplateList
from mailmanclient.restbase.base import RESTObject
from mailmanclient.restbase.page import Page

__metaclass__ = type
__all__ = [
    'MailingList'
]


class MailingList(RESTObject):

    _properties = ('advertised', 'display_name', 'fqdn_listname', 'list_id',
                   'list_name', 'mail_host', 'member_count', 'volume',
                   'self_link', 'description')

    def __init__(self, connection, url, data=None):
        super(MailingList, self).__init__(connection, url, data)
        self._settings = None

    def __repr__(self):
        return '<List {0!r}>'.format(self.fqdn_listname)

    @property
    def owners(self):
        """All MailingList owners."""
        return self.get_roster('owner')

    def get_roster(self, roster, fields=None):
        """Get roster of the MailingList.

        If the fields is specified without `self_link` and `address`, they are
        added since it is required for returning the response.

        :param str roster: One of the Membership rosters from
           'owner', 'moderator', 'member' and 'nonmember'.
        :param List[str] fields: List of Member's fields to fetch from the
           API. Skipping certain fields can speed up the API response
           when they aren't required since they need to be fetched from
           database individually.
        """
        url = self._url + '/roster/{}'.format(roster)
        if fields is not None:
            # We cannot instantiate Member object without address and
            # self_link objects, so just add them too. They don't add
            # a lot of overhead.
            if 'address' not in fields:
                fields.append('address')
            if 'self_link' not in fields:
                fields.append('self_link')
            url += '?' + '&'.join('fields={}'.format(each) for each in fields)
        response, content = self._connection.call(url)
        if 'entries' not in content:
            return []
        else:
            return [Member(self._connection, entry['self_link'], entry)
                    for entry in sorted(content['entries'],
                                        key=itemgetter('address'))]

    @property
    def moderators(self):
        """All MailingList moderators."""
        return self.get_roster('moderator')

    @property
    def members(self):
        """All MailingList members."""
        return self.get_roster('member')

    @property
    def nonmembers(self):
        """All MailingList non-members."""
        return self.get_roster('nonmember')

    def get_member_page(self, count=50, page=1, fields=None):
        """Return a paginated list of MailingList's members.

        :param int count: Count of members in one page.
        :param int page: The page number.
        """
        url = 'lists/{0}/roster/member'.format(self.fqdn_listname)
        return Page(self._connection, url, Member, count, page)

    def find_members(
            self, address=None, role=None, page=None, count=50):
        """Find a Mailinglist's members.

        This provides a filtering API for list's Members including,
        non-members, owners and moderators by speciying the role.

        :param str address: Member's address.
        :param str role: Member's role.
        :param int page: Page number for paginated results.
        :param int count: Number of results per-page for paginated results.

        """
        data = {'list_id': self.list_id}
        if address:
            data['subscriber'] = address
        if role:
            data['role'] = role

        url = 'members/find?{}'.format(urlencode(data, doseq=True))
        if page is None:
            response, content = self._connection.call(url, data)
            if 'entries' not in content:
                return []
            return [Member(self._connection, entry['self_link'], entry)
                    for entry in content['entries']]
        else:
            return Page(self._connection, url, Member, count, page)

    @property
    def settings(self):
        """All MailingList settings."""
        if self._settings is None:
            self._settings = Settings(
                self._connection,
                'lists/{0}/config'.format(self.fqdn_listname))
        return self._settings

    @property
    def held(self):
        """Held messages of a MailingList.."""
        response, content = self._connection.call(
            'lists/{0}/held'.format(self.fqdn_listname), None, 'GET')
        if 'entries' not in content:
            return []
        return [HeldMessage(self._connection, entry['self_link'], entry)
                for entry in content['entries']]

    def get_held_page(self, count=50, page=1):
        """Paginated list of held messages for the MailingList.

        :param int page: Page number for paginated results.
        :param int count: Number of results per-page for paginated results.
        """
        url = 'lists/{0}/held'.format(self.fqdn_listname)
        return Page(self._connection, url, HeldMessage, count, page)

    def get_held_count(self):
        """Get a count of held messages for the MailingList."""
        response, json = self._connection.call(
            'lists/{}/held/count'.format(self.fqdn_listname), None, 'GET')
        return json['count']

    def get_held_message(self, held_id):
        """Get a single held message for MailingList.

        :param int held_id: Held message id to get.
        """
        url = 'lists/{0}/held/{1}'.format(self.fqdn_listname, held_id)
        return HeldMessage(self._connection, url)

    @property
    def requests(self):
        """See :meth:`get_requests`."""
        return self.get_requests()

    @property
    def unsubscription_requests(self):
        """Get a list of subscription requests pending moderator Approval."""
        return self.get_requests(request_type='unsubscription')

    def get_requests(self, token_owner=None, request_type='subscription'):
        """Return a list of dicts with subscription requests.

        This is the new API for requests which allows filtering via
        `token_owner` since it isn't possible to do so via the property
        requests.

        :param token_owner: Who owns the pending requests?  Should be one in
            'no_one', 'moderator' and 'subscriber'.
        :param request_type: The type of pending request. Value should be in
            'subscription' or 'unsubscription'. Defaults to 'subscription'.
        """
        url = 'lists/{0}/requests'.format(self.fqdn_listname)
        fragments = []
        if token_owner:
            fragments.append('token_owner={}'.format(token_owner))
        if request_type:
            fragments.append('request_type={}'.format(request_type))
        if fragments:
            url += '?{}'.format('&'.join(fragments))
        response, content = self._connection.call(url, None, 'GET')
        if 'entries' not in content:
            return []
        else:
            entries = []
            for entry in content['entries']:
                request = dict(email=entry['email'],
                               token=entry['token'],
                               display_name=entry['display_name'],
                               token_owner=entry['token_owner'],
                               list_id=entry['list_id'],
                               request_date=entry['when'])
                entries.append(request)
        return entries

    def get_requests_count(self, token_owner=None):
        """Return a total count of pending subscription requests.

        This should be a faster query when *all* the requests aren't needed and
        only a count is needed to display on the badge in List's settings page.

        :param token_owner: Who owns the pending requests?  Should be one in
            'no_one', 'moderator' and 'subscriber'.
        :returns: The count of pending requests.
        """
        url = 'lists/{}/requests/count'.format(self.fqdn_listname)
        if token_owner:
            url += '?token_owner={}'.format(token_owner)
        response, json = self._connection.call(url)
        return json['count']

    def get_request(self, token):
        """Get an individual pending request for the given token.

        :param token: The token for the request.
        :returns: The request dictionary.
        """
        url = 'lists/{}/requests/{}'.format(self.fqdn_listname, token)
        response, json = self._connection.call(url)
        return json

    @property
    def archivers(self):
        """Get a list of MailingList archivers."""
        url = 'lists/{0}/archivers'.format(self.list_id)
        return ListArchivers(self._connection, url, self)

    @archivers.setter
    def archivers(self, new_value):
        url = 'lists/{0}/archivers'.format(self.list_id)
        archivers = ListArchivers(self._connection, url, self)
        archivers.update(new_value)
        archivers.save()

    def add_owner(self, address, display_name=None):
        """Add a list owner.

        :param str address: Email address of the owner.
        :param str display_name: Display name of the Owner.
        """
        self.add_role('owner', address, display_name)

    def add_moderator(self, address, display_name=None):
        """Add a list moderator.

        :param str address: Email address of the moderator.
        :param str display_name: Display name of the moderator.
        """
        self.add_role('moderator', address, display_name)

    def add_role(self, role, address, display_name=None):
        """Add a new Member with a specific role.

        :param str role: The role for the new member.
        :param str address: A valid email address for the new Member.
        :param str display_name: An optional display name for the Member.
        """
        data = dict(list_id=self.list_id,
                    subscriber=address,
                    display_name=display_name,
                    role=role)
        self._connection.call('members', data)

    def remove_owner(self, address):
        """Remove a list owner.

        :param str address: Email address of the owner to remove.
        """
        self.remove_role('owner', address)

    def remove_moderator(self, address):
        """Remove a list moderator.

        :param str address: Email address of the moderator to remove.
        """
        self.remove_role('moderator', address)

    def remove_role(self, role, address):
        """Remove a list Member with a specific Role.

        :param str role: The role for the new member.
        :param str address: A valid email address for the new Member.
        """
        url = 'lists/%s/%s/%s' % (
            self.fqdn_listname, role, quote_plus(address))
        self._connection.call(url, method='DELETE')

    def moderate_message(self, request_id, action, comment=None):
        """Moderate a held message.

        :param request_id: Id of the held message.
        :type request_id: Int.
        :param action: Action to perform on held message.
        :type action: String.
        :param comment: The reason for action, only supported for rejection.
        :type comment: str
        """
        data = dict(action=action)
        if comment is not None:
            data['comment'] = comment

        path = 'lists/{0}/held/{1}'.format(
            self.fqdn_listname, str(request_id))
        response, content = self._connection.call(
            path, data, 'POST')
        return response

    def discard_message(self, request_id):
        """Shortcut for moderate_message.

        :param str request_id: The request_id of the held message.
        """
        return self.moderate_message(request_id, 'discard')

    def reject_message(self, request_id, reason=None):
        """Shortcut for moderate_message.

        :param str request_id: The request_id of the held message.
        :param str reason: An optional reason for rejection of the message.
        """
        return self.moderate_message(request_id, 'reject', reason)

    def defer_message(self, request_id):
        """Shortcut for moderate_message.

        :param str request_id: The request_id of the held message.
        """
        return self.moderate_message(request_id, 'defer')

    def accept_message(self, request_id):
        """Shortcut for moderate_message.

        :param str request_id: The request_id of the held message.
        """
        return self.moderate_message(request_id, 'accept')

    def moderate_request(self, request_id, action, reason=None):
        """
        Moderate a subscription request.

        :param action: accept|reject|discard|defer
        :type action: str.
        :param reason: The reason associated with rejections.
        :type reason: str
        """
        path = 'lists/{0}/requests/{1}'.format(self.list_id, request_id)
        data = {'action': action}
        if reason:
            data['reason'] = reason
        response, content = self._connection.call(path, data)
        return response

    def manage_request(self, token, action):
        """Alias for moderate_request, kept for compatibility"""
        warnings.warn(
            'The `manage_request()` method has been replaced by '
            '`moderate_request()` and will be removed in the future.',
            DeprecationWarning, stacklevel=2)
        return self.moderate_request(token, action)

    def accept_request(self, request_id):
        """Shortcut to accept a subscription request."""
        return self.moderate_request(request_id, 'accept')

    def reject_request(self, request_id):
        """Shortcut to reject a subscription request."""
        return self.moderate_request(request_id, 'reject')

    def discard_request(self, request_id):
        """Shortcut to discard a subscription request."""
        return self.moderate_request(request_id, 'discard')

    def defer_request(self, request_id):
        """Shortcut to defer a subscription request."""
        return self.moderate_request(request_id, 'defer')

    def _get_membership(self, email, role):
        """Get a single membership resource.

        :param address: The email address of the member for this list.
        :param role: The membership role.
        :return: A member proxy object.
        """
        # In order to get the member object we query the REST API for
        # the member. Incase there is no matching subscription, an
        # HTTPError is returned instead.
        try:
            path = 'lists/{0}/{1}/{2}'.format(
                self.list_id, role, quote_plus(email))
            response, content = self._connection.call(path)
            return Member(self._connection, content['self_link'], content)
        except HTTPError:
            raise ValueError('%s is not a %s address of %s' %
                             (email, role, self.fqdn_listname))

    def get_member(self, email):
        """Get a Member of the list.

        :param address: The email address of the member for this list.
        :return: A member proxy object.
        """
        return self._get_membership(email, 'member')

    def get_nonmember(self, email):
        """Get a non-member of the list.

        :param address: The email address of the non-member for this list.
        :return: A member proxy object.
        """
        return self._get_membership(email, 'nonmember')

    def subscribe(self, address, display_name=None, pre_verified=False,
                  pre_confirmed=False, pre_approved=False, invitation=False,
                  send_welcome_message=None, delivery_mode=None,
                  delivery_status=None):
        """Subscribe an email address to a mailing list.

        :param address: Email address to subscribe to the list.
        :type address: str
        :param display_name: The real name of the new member.
        :type display_name: str
        :param pre_verified: True if the address has been verified.
        :type pre_verified: bool
        :param pre_confirmed: True if membership has been approved by the user.
        :type pre_confirmed: bool
        :param pre_approved: True if membership is moderator-approved.
        :type pre_approved: bool
        :param invitation: True if this is an invitation to join the list.
        :type invitation: bool
        :param send_welcome_message: True if welcome message should be sent.
        :type send_welcome_message: bool
        :param delivery_mode: Delivery mode of the Member.
        :type delivery_mode: str. One between 'regular', 'plaintext_digests',
            'mime_digests', 'summary_digests'.
        :param delivery_status: Delivery status of the Member.
        :type delivery_status: str. One between 'enabled', 'by_owner',
            'by_moderator', 'by_user'.
        :return: A member proxy object.
        """
        data = dict(
            list_id=self.list_id,
            subscriber=address,
        )
        if display_name:
            data['display_name'] = display_name
        if pre_verified:
            data['pre_verified'] = True
        if pre_confirmed:
            data['pre_confirmed'] = True
        if pre_approved:
            data['pre_approved'] = True
        if invitation:
            data['invitation'] = True
        if delivery_mode:
            data['delivery_mode'] = delivery_mode
        if delivery_status:
            data['delivery_status'] = delivery_status
        # Even if it is False, we should send this value because it means we
        # should suppress welcome message, so check for None value to skip the
        # parameter.
        if send_welcome_message is not None:
            data['send_welcome_message'] = send_welcome_message
        response, content = self._connection.call('members', data)
        # If a member is not immediately subscribed (i.e. verificatoin,
        # confirmation or approval need), the response content is returned.
        if response.status_code == 202:
            return content
        # If the subscription is executed immediately, a member object
        # is returned.
        return Member(self._connection, response.headers.get('location'))

    def unsubscribe(self, email, pre_confirmed=None,
                    pre_approved=None):
        """Unsubscribe an email address from a mailing list.

        :param address: Email address to unsubscribe.
        :type address: str
        :param pre_confirmed: True if unsubscribe is approved by the user.
        :type pre_confirmed: bool
        :param pre_approved: True if unsubscribe is moderator-approved.
        :type pre_approved: bool
        """
        data = dict()
        if pre_confirmed is not None:
            data['pre_confirmed'] = pre_confirmed
        if pre_approved is not None:
            data['pre_approved'] = pre_approved
        try:
            path = 'lists/{0}/member/{1}'.format(self.list_id, email)
            response, json = self._connection.call(path, data, method='DELETE')
            if response.status_code == 202:
                return json
        except HTTPError:
            # The member link does not exist, i.e. they are not a member
            raise ValueError('%s is not a member address of %s' %
                             (email, self.fqdn_listname))

    def mass_unsubscribe(self, email_list):
        """Unsubscribe a list of emails from a mailing list.

        This function return a json of emails mapped to booleans based
        on whether they were unsubscribed or not, for whatever reasons

        :param email_list: list of emails to unsubscribe
        """
        try:
            path = 'lists/{}/roster/member'.format(self.list_id)
            response, content = self._connection.call(
                path, {'emails': email_list}, 'DELETE')
            return content
        except HTTPError as e:
            raise ValueError(str(e))

    @property
    def bans(self):
        """A list of banned addresses for this MailingList."""
        from mailmanclient.restobjects.ban import Bans
        url = 'lists/{0}/bans'.format(self.list_id)
        return Bans(self._connection, url, mlist=self)

    def get_bans_page(self, count=50, page=1):
        """Get a paginated list of bans for this MailingList.

        :param int page: Page number for paginated results.
        :param int count: Number of results per-page for paginated results.
        """
        from mailmanclient.restobjects.ban import BannedAddress
        url = 'lists/{0}/bans'.format(self.list_id)
        return Page(self._connection, url, BannedAddress, count, page)

    @property
    def header_matches(self):
        """A list of header-match rules for the MailingList."""
        url = 'lists/{0}/header-matches'.format(self.list_id)
        return HeaderMatches(self._connection, url, self)

    @property
    def templates(self):
        """Get a list of MailingList templates."""
        url = self._url + '/uris'
        return TemplateList(self._connection, url)

    def set_template(self, template_name, uri, username=None, password=None):
        """Set a MailingList template URI.

        :param str template_name: The name of the template.
        :param str uri: The URI to fetch the template.
        :param str username: Username for fetching template from uri.
        :param str password: Password for fetching template from uri.
        """
        url = self._url + '/uris'
        data = {template_name: uri}
        if username is not None and password is not None:
            data['username'] = username
            data['password'] = password
        return self._connection.call(url, data, 'PATCH')[1]

    def _check_membership(self, address, allowed_roles):
        """
        Given an address and role, check if there is a membership record that
        matches the given address with a given role for this Mailing List.
        """
        url = 'members/find'
        data = {'subscriber': address,
                'list_id': self.list_id}
        response, content = self._connection.call(url, data=data)
        if 'entries' not in content:
            return False
        for membership in content['entries']:
            # We check for all the returned roles for this User and MailingList
            if membership['role'] in allowed_roles:
                return True
        return False

    def is_owner(self, address):
        """
        Given an address, checks if the given address is an owner of this
        mailing list.
        """
        return self._check_membership(address=address,
                                      allowed_roles=('owner',))

    def is_moderator(self, address):
        """
        Given an address, checks if the given address is a moderator of this
        mailing list.
        """
        return self._check_membership(address=address,
                                      allowed_roles=('moderator',))

    def is_member(self, address):
        """
        Given an address, checks if the given address is subscribed to this
        mailing list.
        """
        return self._check_membership(address=address,
                                      allowed_roles=('member',))

    def is_owner_or_mod(self, address):
        """
        Given an address, checks if the given address is either a owner or
        a moderator of this list.

        It is possible for them to be both owner and moderator.
        """
        return self._check_membership(address=address,
                                      allowed_roles=('owner', 'moderator'))



import sys
from typing import Mapping, Any, Tuple, NewType
if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    # mypy understnads version_info check better than ImportError. So for the
    # sake of static typing checks, we are for once going to do LBYL.
    from typing_extensions import Protocol

__all__ = [
    'HTTPClientProto',
    'ResponseType',
    'ContentType',
    'ConnectionProto',
]


ContentType = NewType('ContentType', Mapping[str, Any])


class ResponseType(Protocol):
    status_code: int
    content: ContentType


class HTTPClientProto(Protocol):
    """A type checking protocol for the HTTP client argument for AsyncClient"""
    async def request(self) -> ResponseType:
        ...


class ConnectionProto(Protocol):
    """A type checking procol for connection object to avoid circular import"""
    async def call(self,
                   path: str,
                   data: Mapping[str, Any] = None,
                   method: str = 'GET') -> Tuple[ResponseType, ContentType]:
        ...



#! /usr/bin/env python3

import os
import re
import sys
import stat
import datetime


FSF = 'by the Free Software Foundation, Inc.'
this_year = datetime.date.today().year
pyre_c = re.compile(r'# Copyright \(C\) ((?P<start>\d{4})-)?(?P<end>\d{4})')
pyre_n = re.compile(r'# Copyright ((?P<start>\d{4})-)?(?P<end>\d{4})')
new_c = '# Copyright (C) {}-{} {}'
new_n = '# Copyright {}-{} {}'

MODE = (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)


if '--noc' in sys.argv:
    pyre = pyre_n
    new = new_n
    sys.argv.remove('--noc')
else:
    pyre = pyre_c
    new = new_c


def do_file(path, owner):
    permissions = os.stat(path).st_mode & MODE
    with open(path) as in_file, open(path + '.out', 'w') as out_file:
        try:
            for line in in_file:
                mo_c = pyre_c.match(line)
                mo_n = pyre_n.match(line)
                if mo_c is None and mo_n is None:
                    out_file.write(line)
                    continue
                mo = (mo_n if mo_c is None else mo_c)
                start = (mo.group('end')
                         if mo.group('start') is None
                         else mo.group('start'))
                if int(start) == this_year:
                    out_file.write(line)
                    continue
                print(new.format(start, this_year, owner), file=out_file)
                print('=>', path)
                for line in in_file:
                    out_file.write(line)
        except UnicodeDecodeError:
            print('Cannot convert path:', path)
            os.remove(path + '.out')
            return
    os.rename(path + '.out', path)
    os.chmod(path, permissions)


def remove(dirs, path):
    try:
        dirs.remove(path)
    except ValueError:
        pass


def do_walk():
    try:
        owner = sys.argv[1]
    except IndexError:
        owner = FSF
    for root, dirs, files in os.walk('.'):
        if root == '.':
            remove(dirs, '.git')
            remove(dirs, '.tox')
            remove(dirs, 'bin')
            remove(dirs, 'contrib')
            remove(dirs, 'develop-eggs')
            remove(dirs, 'eggs')
            remove(dirs, 'parts')
            remove(dirs, 'gnu-COPYING-GPL')
            remove(dirs, '.installed.cfg')
            remove(dirs, '.bzrignore')
            remove(dirs, 'distribute_setup.py')
        if root == './src':
            remove(dirs, 'mailman.egg-info')
        if root == './src/mailman':
            remove(dirs, 'messages')
        for file_name in files:
            if os.path.splitext(file_name)[1] in ('.pyc', '.gz', '.egg'):
                continue
            path = os.path.join(root, file_name)
            if os.path.isfile(path):
                do_file(path, owner)


if __name__ == '__main__':
    do_walk()



import httpx
import pytest
import pytest_asyncio
import concurrent.futures

from mailmanclient.asynclient import AsyncClient
from mailmanclient import Client


if pytest_asyncio.__version__ < '0.17':
    pytest_asyncio.fixture = pytest.fixture


@pytest_asyncio.fixture(autouse=True)
def setup():
    """Setup for testing. Create test data."""
    client = Client('http://localhost:9001/3.1', 'restadmin', 'restpass')
    print('Loading test data...')
    try:
        domain = client.create_domain('example.com')
    except Exception:
        domain = client.get_domain('example.com')
    # Create some lists.
    lists = []
    print('Creating lists...')
    with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
        future_to_id = {
            executor.submit(domain.create_list, f'list{i}'): f'list{i}'
            for i in range(10)}
        for future in concurrent.futures.as_completed(future_to_id):
            lists.append(future.result())

    # Subscribe some addresses.
    print('Creating subscirbers...')
    for i, ml in enumerate(lists):
        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            futures = [
                executor.submit(
                    ml.subscribe, f'mylist{each}@example.com',
                    pre_verified=True, pre_confirmed=True, pre_approved=True)
                for each in range(i*5)]
            for _future in concurrent.futures.as_completed(futures):
                print('.', end='')
    yield
    # Cleanup after test.
    for ml in lists:
        ml.delete()
    domain.delete()


@pytest_asyncio.fixture
async def client():
    async with httpx.AsyncClient() as conn:
        client = AsyncClient(
            conn, 'http://localhost:9001/3.1', 'restadmin', 'restpass')
        yield client


@pytest.mark.asyncio
async def test_async_client(client):
    domains = await client.domains()
    for each in domains:
        print(f'Domain: {each.mail_host}')


@pytest.mark.asyncio
async def test_get_lists(client):
    lists = await client.lists()
    for ml in lists:
        print(f'Mailinglist: {ml.fqdn_listname}')
        await ml.config()


@pytest.mark.asyncio
async def test_get_members(client):
    members = await client.members()
    for member in members:
        print(f'Member: {member.role} {member.list_id} {member.email} ',
              end='')


@pytest.mark.asyncio
async def test_get_users(client):
    users = await client.users()
    for user in users:
        print(f'User: {user.user_id} ', end='')
        addrs = await user.addresses()
        for addr in addrs:
            print(f'UserAddress: {addr.email}', end='')
        prefs = await user.preferences()
        print('UserPreference: {}'.format(prefs._data), end='')


@pytest.mark.asyncio
async def test_get_addresses(client):
    addresses = await client.addresses()
    for addr in addresses:
        print(f'Address: {addr.email}', end='')



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Async domain object."""

__all__ = [
    'Domain'
]

from mailmanclient.restbase.async_base import RESTObject


class Domain(RESTObject):

    _properties = ('alias_domain', 'description', 'mail_host', 'self_link')

    def __repr__(self) -> str:
        return '<Domain {}>'.format(self.mail_host)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
import warnings

from mailmanclient.restobjects.mailinglist import MailingList
from mailmanclient.restobjects.templates import TemplateList
from mailmanclient.restobjects.user import User
from mailmanclient.restbase.base import RESTObject
from mailmanclient.restbase.page import Page

__metaclass__ = type
__all__ = [
    'Domain'
]


class Domain(RESTObject):

    _properties = ('alias_domain', 'description', 'mail_host', 'self_link')

    def __repr__(self):
        return '<Domain {0!r}>'.format(self.mail_host)

    def __str__(self):
        return self.mail_host

    @property
    def web_host(self):
        warnings.warn(
            'The `Domain.web_host` attribute is deprecated. It is not used '
            'any more and will be removed in the future.',
            DeprecationWarning, stacklevel=2)
        return 'http://{}'.format(self.mail_host)

    @property
    def base_url(self):
        warnings.warn(
            'The `Domain.base_url` attribute is deprecated. It is not used '
            'any more and will be removed in the future.',
            DeprecationWarning, stacklevel=2)
        return 'http://{}'.format(self.mail_host)

    @property
    def owners(self):
        url = self._url + '/owners'
        response, content = self._connection.call(url)
        if 'entries' not in content:
            return []
        else:
            return [User(self._connection, entry['self_link'], entry)
                    for entry in content['entries']]

    @property
    def lists(self):
        return self.get_lists()

    def get_lists(self, advertised=None):
        url = 'domains/{0}/lists'.format(self.mail_host)
        if advertised:
            url += '?advertised=true'
        response, content = self._connection.call(url)
        if 'entries' not in content:
            return []
        return [MailingList(self._connection, entry['self_link'], entry)
                for entry in content['entries']]

    def get_list_page(self, count=50, page=1, advertised=None):
        url = 'domains/{0}/lists'.format(self.mail_host)
        if advertised:
            url += '?advertised=true'
        return Page(self._connection, url, MailingList, count, page)

    def create_list(self, list_name, style_name=None):
        fqdn_listname = '{0}@{1}'.format(list_name, self.mail_host)
        data = dict(fqdn_listname=fqdn_listname)
        if style_name is not None:
            data['style_name'] = style_name
        response, content = self._connection.call('lists', data)
        return MailingList(self._connection, response.headers.get('location'))

    # TODO: Add this when the API supports removing a single owner.
    # def remove_owner(self, owner):
    #     url = self._url + '/owners/{}'.format(owner)
    #     response, content = self._connection.call(
    #         url, method='DELETE')
    #     return response

    def remove_all_owners(self):
        url = self._url + '/owners'
        response, content = self._connection.call(
            url, method='DELETE')
        return response

    def add_owner(self, owner):
        url = self._url + '/owners'
        response, content = self._connection.call(
            url, {'owner': owner})

    @property
    def templates(self):
        url = self._url + '/uris'
        return TemplateList(self._connection, url)

    def set_template(self, template_name, uri, username=None, password=None):
        url = self._url + '/uris'
        data = {template_name: uri}
        if username is not None and password is not None:
            data['username'] = username
            data['password'] = password
        return self._connection.call(url, data, 'PATCH')[1]



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Testing utilities."""

import pytest
import subprocess
import socket
from contextlib import closing


def check_core():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        if sock.connect_ex(('localhost', 9001)) == 0:
            return True
        else:
            return False


@pytest.fixture(scope='module', autouse=True)
def mailman_core(request, watcher_getter):
    """Mailman Core instance which is ready to be used by the tests"""
    def teardown_core():
        print('Stopping Mailman Server')
        subprocess.run(['mailman', 'stop'])
        subprocess.run(['rm', '-rf', 'var/'])
    request.addfinalizer(teardown_core)

    print('Starting Mailman Server')

    return watcher_getter(
        name='master',
        arguments=['-C', 'mailman_test.cfg'],
        checker=check_core,
        request=request,
    )



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Base classes for async objects."""

__all__ = [
    'RESTBase',
    'RESTObject',
]

from typing import Sequence, Any, Tuple
from mailmanclient.restobjects.types import (
    ConnectionProto, ResponseType, ContentType)


class RESTBase:

    _properties: Sequence[str] = ['self_link']
    _writable_properties: Sequence[str] = []
    _read_only_properties: Sequence[str] = ['self_link']

    def __init__(self,
                 connection: ConnectionProto, data: ContentType,
                 ) -> None:
        """
        :param connection: API connection object to fetch sub-resources.
        :param data: Data from REST API.
        """
        self._data = data
        self._connection = connection
        self._changed_rest_data = {}

    def __repr__(self) -> str:
        """Provide a default repr for all object types."""
        return '<{} at {}>'.format(self.__class__, self._data.get('self_link'))

    def _get(self, key: str) -> str:
        """Get the value of 'key' from object's REST data.

        :param key: The key to get value for.
        """
        if self._properties is not None:
            # Some REST key/values may not be returned by Mailman if the value
            # is None.
            if key in self._data:
                return self._data.get(key)
            raise KeyError(key)
        else:
            return self._data.get(key)

    def _set(self, key: str, value: Any) -> None:
        if (key in self._read_only_properties or (
                self._writable_properties is not None
                and key not in self._writable_properties)):
            raise ValueError(f'{key} is read-only')
        # Don't check that the key is in _properties, the accepted values for
        # write may be different from the returned values (eg: User.password
        # and User.cleartext_password).
        if self._data.get(key, None) == value:
            return
        self._changed_rest_data[key] = value

    def _reset_cache(self) -> None:
        self._changed_rest_data = {}
        self._data = None

    async def save(self) -> Tuple[ResponseType, ContentType]:
        res = await self._connection.call(
            self._data.get('self_link'),
            self._changed_rest_data,
            method='PATCH'
            )
        self._reset_cache()
        return res


class RESTObject(RESTBase):

    def __getattr__(self, name) -> Any:
        try:
            return self._get(name)
        except KeyError:
            raise AttributeError(
                '"{}" has no attribute "{}"'.format(
                    self.__class__.__name__, name))

    def __setattr__(self, name: str, value: Any):
        if self._properties and (name not in self._properties):
            return super().__setattr__(name, value)
        return self._set(name, value)

    async def delete(self) -> Tuple[ResponseType, ContentType]:
        res = await self.connection.call(
            self._data.get('self_link'), method='DELETE')
        self._reset_cache()
        return res



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Async Mailinglist object.."""

__all__ = [
    'MailingList',
]

from enum import Enum
from typing import List
from mailmanclient.restbase.async_base import RESTObject
from mailmanclient.asyncobjects.member import Member
from mailmanclient.restobjects.utils import list_of_objects
from mailmanclient.restobjects.types import ConnectionProto, ContentType


class MemberRole(Enum):
    """Member role values."""
    member = 'member'
    owner = 'owner'
    moderator = 'moderator'
    nonmember = 'nonmember'


class MailingList(RESTObject):

    _properties = ('advertised', 'display_name', 'fqdn_listname', 'list_id',
                   'list_name', 'mail_host', 'member_count', 'volume',
                   'self_link', 'description')

    def __repr__(self) -> str:
        return '<MailingList {}>'.format(self.fqdn_listname)

    async def config(self) -> 'Config':
        """Get MailingList settings.

        /<api>/lists/<listid>/config
        """
        path = 'lists/{}/config'.format(self.list_id)
        _, content = await self._connection.call(path)
        return Config(self, self._connection, content)

    async def get_roster(self, role) -> List[Member]:
        """Get MailingList roster.

        /<api>/lists/<listid>/roster/<role>
        """
        path = 'lists/{}/roster/{}'.format(self.fqdn_listname, role)
        _, content = await self._connection.call(path)
        return list_of_objects(Member, content, self._connection)

    async def members(self) -> List[Member]:
        """Get Mailinglist members (subscribers.)

        /<api>/lists/<listid>/roster/member
        """
        return await self.get_roster(MemberRole.member)

    async def owners(self) -> List[Member]:
        """Get Mailinglist owners.

        /<api>/lists/<listid>/roster/owner
        """
        return await self.get_roster(MemberRole.owner)

    async def moderators(self) -> List[Member]:
        """Get Mailinglist moderators.

        /<api>/lists/<listid>/roster/moderator
        """
        return await self.get_roster(MemberRole.moderator)

    async def nonmember(self) -> List[Member]:
        """Get Mailinglist nonmembers.

        /<api>/lists/<listid>/roster/nonmember
        """
        return await self.get_roster(MemberRole.nonmember)


class Config(RESTObject):

    _read_only_properties = (
        'bounces_address',
        'created_at',
        'digest_last_sent_at',
        'fqdn_listname',
        'join_address',
        'last_post_at',
        'leave_address',
        'list_id',
        'list_name',
        'mail_host',
        'next_digest_number',
        'no_reply_address',
        'owner_address',
        'post_id',
        'posting_address',
        'request_address',
        'scheme',
        'self_link',
        'volume',
        'web_host',
        )

    def __init__(self,
                 mailing_list: MailingList,
                 connection: ConnectionProto,
                 data: ContentType) -> None:
        super().__init__(connection, data)
        self.mailing_list = mailing_list

    def __repr__(self) -> str:
        return '<Settings for {}>'.format(self.mailing_list.fqdn_listname)



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Old module for backwards compatibility"""

from __future__ import absolute_import, print_function, unicode_literals

from mailmanclient import *                                      # noqa

__metaclass__ = type

# XXX: This module exists for backwards compatibility with older versions of
# MailmanClient which had all the API under this module. It will be removed in
# future after few cycles of deprecation.



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
#
# flake8: noqa

"""Package contents."""

from mailmanclient.client import Client
from mailmanclient.constants import __version__
from mailmanclient.restbase.connection import MailmanConnectionError
from mailmanclient.restobjects.address import Address, Addresses
from mailmanclient.restobjects.ban import Bans, BannedAddress
from mailmanclient.restobjects.configuration import Configuration
from mailmanclient.restobjects.domain import Domain
from mailmanclient.restobjects.header_match import HeaderMatch, HeaderMatches
from mailmanclient.restobjects.held_message import HeldMessage
from mailmanclient.restobjects.archivers import ListArchivers
from mailmanclient.restobjects.mailinglist import MailingList
from mailmanclient.restobjects.member import Member
from mailmanclient.restobjects.preferences import Preferences, PreferencesMixin
from mailmanclient.restobjects.queue import Queue
from mailmanclient.restobjects.settings import Settings
from mailmanclient.restobjects.user import User


__metaclass__ = type
__all__ = [
    'Address',
    'Addresses',
    'Bans',
    'BannedAddress',
    'Client',
    'Configuration',
    'Domain'
    'HeaderMatch',
    'HeaderMatches',
    'HeldMessage',
    'ListArchivers',
    'MailingList',
    'MailmanConnectionError',
    'Member',
    'Preferences',
    'PreferencesMixin',
    'Queue',
    'Settings',
    'User',
    '__version__',
]


__all__ = [bytes(x, 'utf-8') for x in __all__]



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""pytest conftest"""

__metaclass__ = type

import os
import pytest


@pytest.fixture
def vcr_cassette_path(request, vcr_cassette_name):
    return os.path.join('src/mailmanclient/tests/data', vcr_cassette_name)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from urllib.error import HTTPError
from urllib.parse import urljoin, urlencode, urlparse, urlunparse

from requests import request

from mailmanclient.constants import __version__

__metaclass__ = type
__all__ = [
    'MailmanConnectionError',
    'Connection'
]


class MailmanConnectionError(Exception):
    """Custom Exception to catch connection errors."""


class Connection:
    """A connection to the REST client."""

    def __init__(self, baseurl, name=None, password=None, request_hooks=None):
        """Initialize a connection to the REST API.

        :param baseurl: The base url to access the Mailman 3 REST API.
        :param name: The Basic Auth user name.  If given, the `password` must
            also be given.
        :param password: The Basic Auth password.  If given the `name` must
            also be given.
        :param request_hooks: A list of callables that can receive the request
            parameters and return them with some changes or unchanged.
        """
        if baseurl[-1] != '/':
            baseurl += '/'
        self.baseurl = baseurl
        self.name = name
        self.password = password
        if name is not None and password is None:
            raise TypeError('`password` is required when `name` is given')
        if name is None and password is not None:
            raise TypeError('`name` is required when `password` is given')
        if name is None:
            self.auth = None
        else:
            self.auth = (name, password)
        self.request_hooks = request_hooks

    def add_hooks(self, request_hooks):
        """Add a list of hooks to an existing connection object.

        :param request_hooks: A list of Request hook which receive the request
            parameters.
        :type request_hooks: List[callables]
        """
        if self.request_hooks is None:
            self.request_hooks = request_hooks
        else:
            self.request_hooks.extend(request_hooks)

    def rewrite_url(self, url):
        """rewrite url component with self.baseurl prefix "scheme://netloc"

        :param url: the URL to rewrite
        :type url: str
        :return: modified URL
        :rtype: str
        """
        # rewrite url component with self.baseurl prefix "scheme://netloc"
        pbaseurl = urlparse(self.baseurl)
        parsed = urlparse(url)
        parsed = parsed._replace(scheme=pbaseurl.scheme,
                                 netloc=pbaseurl.netloc)
        return urlunparse(parsed)

    def _process_request_hooks(self, params):
        """Given the request parameters, pass them through the list of hooks.

        Hooks are simple callables that are provided with request parameters
        and return the same parameters, possibly with some modification or not.

        :param params: The HTTP request parameters.
        :returns: The HTTP request parameters.
        """
        for hook in self.request_hooks:
            try:
                params = hook(params)
            except Exception:
                print('[DEBUG] Failed to run hook {hook}')
        return params

    def _prepare_request(self, path, data, method):
        headers = {
            'User-Agent': 'GNU Mailman REST client v{0}'.format(__version__),
            }
        data_str = None
        if data is not None:
            data_str = urlencode(data, doseq=True, encoding='utf-8')
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
        if method is None:
            if data_str is None:
                method = 'GET'
            else:
                method = 'POST'
        method = method.upper()
        url = urljoin(self.baseurl, path)
        url = self.rewrite_url(url)
        return dict(url=url, method=method, data=data_str,
                    headers=headers)

    def call(self, path, data=None, method=None):
        """Make a call to the Mailman REST API.

        :param path: The url path to the resource.
        :type path: str
        :param data: Data to send, implies POST (default) or PUT.
        :type data: dict
        :param method: The HTTP method to call.  Defaults to GET when `data`
            is None or POST if `data` is given.
        :type method: str
        :return: The response content, which will be None, a dictionary, or a
            list depending on the actual JSON type returned.
        :rtype: None, list, dict
        :raises HTTPError: when a non-2xx status code is returned.
        """
        params = self._prepare_request(path, data, method)
        if self.request_hooks:
            params = self._process_request_hooks(params)

        try:
            response = request(**params, auth=self.auth)
            # content = response.content
            # If we did not get a 2xx status code, make this look like a
            # urllib2 exception, for backward compatibility.
            if response.status_code // 100 != 2:
                try:
                    err = response.json()
                    # If this fails, a ValueError is raised. It means either
                    # the response is malformed JSON or None.
                    error_msg = err['description']
                    # This can fail if the error message does not container
                    # description field.
                except (KeyError, ValueError):
                    error_msg = response.text

                raise HTTPError(params.get('url'), response.status_code,
                                error_msg, response, None)
            if len(response.content) == 0:
                return response, None
            return response, response.json()
        except HTTPError:
            raise
        except IOError as e:
            raise MailmanConnectionError(
                'Could not connect to Mailman API: ', repr(e))



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Member async object."""

__all__ = [
    'Member',
]

from mailmanclient.restbase.async_base import RESTObject
from mailmanclient.asyncobjects.preferences import PreferencesMixin


class Member(RESTObject, PreferencesMixin):

    _properties = ('address', 'delivery_mode', 'email', 'list_id',
                   'moderation_action', 'display_name', 'role', 'self_link',
                   'subscription_mode', 'member_id')
    _writable_properties = ('address', 'delivery_mode', 'moderation_action')

    def __repr__(self):
        return '<Member {0!r} on {1!r} with role {2!r}>'.format(
            self.email, self.list_id, self.role)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

from collections.abc import MutableMapping, Sequence

__metaclass__ = type
__all__ = [
    'RESTBase',
    'RESTDict',
    'RESTList',
    'RESTObject'
]


class RESTBase:
    """
    Base class for data coming from the REST API.

    Subclasses can (and sometimes must) define some attributes to handle a
    particular entity.

    :cvar _properties: the list of expected entity properties. This is required
      for API elements that behave like an object, with REST data accessed
      through attributes. If this value is None, the REST data is used to
      list available properties.
    :cvar _writable_properties: list of properties that can be written to using
        a `PATCH` request. If this value is `None`, all properties are
        writable.
    :cvar _read_only_properties: list of properties that cannot be written to
      (defaults to `self_link` only).
    :cvar _autosave: automatically send a `PATCH` request to the API when a
        value is changed. Otherwise, the `save()` method must be called.
    """

    _properties = None
    _writable_properties = None
    _read_only_properties = ['self_link']
    _autosave = False

    def __init__(self, connection, url, data=None):
        """
        :param connection: An API connection object.
        :type connection: Connection.
        :param url: The url of the API endpoint.
        :type url: str.
        :param data: The initial data to use.
        :type data: dict.
        """
        self._connection = connection
        self._url = url
        self._rest_data = data
        self._changed_rest_data = {}

    def __repr__(self):
        return '<{0} at {1}>'.format(self.__class__.__name__, self._url)

    @property
    def rest_data(self):
        """Get data from API and cache it (only once per instance)."""
        if self._rest_data is None:
            response, content = self._connection.call(self._url)
            if isinstance(content, dict) and 'http_etag' in content:
                del content['http_etag']  # We don't care about etags.
            self._rest_data = content
        return self._rest_data

    def _get(self, key):
        if self._properties is not None:
            # Some REST key/values may not be returned by Mailman if the value
            # is None.
            if key in self._properties:
                return self.rest_data.get(key)
            raise KeyError(key)
        else:
            return self.rest_data[key]

    def _set(self, key, value):
        if (key in self._read_only_properties or (
                self._writable_properties is not None
                and key not in self._writable_properties)):
            raise ValueError('value is read-only')
        # Don't check that the key is in _properties, the accepted values for
        # write may be different from the returned values (eg: User.password
        # and User.cleartext_password).
        if key in self.rest_data and self.rest_data[key] == value:
            return  # Nothing to do
        self._changed_rest_data[key] = value
        if self._autosave:
            self.save()

    def _reset_cache(self):
        self._changed_rest_data = {}
        self._rest_data = None

    def save(self):
        response, content = self._connection.call(
            self._url, self._changed_rest_data, method='PATCH')
        self._reset_cache()


class RESTObject(RESTBase):
    """Base class for REST data that behaves like an object with attributes."""

    def __getattr__(self, name):
        try:
            return self._get(name)
        except KeyError:
            # Transform the KeyError into the more appropriate AttributeError
            raise AttributeError(
                "'{0}' object has no attribute '{1}'".format(
                    self.__class__.__name__, name))

    def __setattr__(self, name, value):
        # RESTObject must list REST-specific properties or we won't be able to
        # store the _connection, _url, etc.
        assert self._properties is not None
        if name not in self._properties:
            return super(RESTObject, self).__setattr__(name, value)
        return self._set(name, value)

    def delete(self):
        self._connection.call(self._url, method='DELETE')
        self._reset_cache()


class RESTDict(RESTBase, MutableMapping):
    """Base class for REST data that behaves like a dictionary."""

    def __repr__(self):
        return repr(self.rest_data)

    def __getitem__(self, key):
        return self._get(key)

    def __setitem__(self, key, value):
        self._set(key, value)

    def __delitem__(self, key):
        raise NotImplementedError("REST dictionnary keys can't be deleted.")

    def __iter__(self):
        for key in self.rest_data:
            if self._properties is not None and key not in self._properties:
                continue
            yield key

    def __len__(self):
        return len(self.rest_data)

    def get(self, key, default=None):
        return self.rest_data.get(key, default)

    def keys(self):
        return list(self)

    def update(self, other):
        # Optimize the update to call save() only once
        _old_autosave = self._autosave
        self._autosave = False
        super(RESTDict, self).update(other)
        self._autosave = _old_autosave
        if self._autosave:
            self.save()


class RESTList(RESTBase, Sequence):
    """
    Base class for REST data that behaves like a list.

    The `_factory` attribute is a callable that will be applied on each
    returned member of the list.
    """

    _factory = lambda x: x  # noqa: E731

    @property
    def rest_data(self):
        if self._rest_data is None:
            response, content = self._connection.call(self._url)
            if 'entries' not in content:
                self._rest_data = []
            else:
                self._rest_data = content['entries']
        return self._rest_data

    def __repr__(self):
        return repr(self.rest_data)

    def __getitem__(self, key):
        return self._factory(self.rest_data[key])

    def __delitem__(self, key):
        self[key].delete()
        self._reset_cache()

    def __len__(self):
        return len(self.rest_data)

    def __iter__(self):
        for entry in self.rest_data:
            yield self._factory(entry)

    def clear(self):
        self._connection.call(self._url, method='DELETE')
        self._reset_cache()



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

from urllib.error import HTTPError
from urllib.parse import quote_plus

from mailmanclient.restobjects.mailinglist import MailingList
from mailmanclient.restbase.base import RESTList, RESTObject

__metaclass__ = type
__all__ = [
    'Bans',
    'BannedAddress'
]


class Bans(RESTList):
    """
    The list of banned addresses from a mailing-list or from the whole site.
    """

    def __init__(self, connection, url, data=None, mlist=None):
        """
        :param mlist: The corresponding list object, or None if it is a global
            ban list.
        :type mlist: MailingList or None.
        """
        super(Bans, self).__init__(connection, url, data)
        self._mlist = mlist
        self._factory = lambda data: BannedAddress(
            self._connection, data['self_link'], data)

    def __repr__(self):
        if self._mlist is None:
            return '<Global bans>'
        else:
            return '<Bans on {0!r}>'.format(self._mlist.list_id)

    def __contains__(self, item):
        # Accept email addresses and BannedAddress restobjects
        if isinstance(item, BannedAddress):
            item = item.email
        if self._rest_data is not None:
            return item in [data['email'] for data in self._rest_data]
        else:
            # Avoid getting the whole list just to check membership
            try:
                response, content = self._connection.call(
                    '{}/{}'.format(self._url, quote_plus(item)))
            except HTTPError as e:
                if e.code == 404:
                    return False
                else:
                    raise
            else:
                return True

    def add(self, email):
        response, content = self._connection.call(self._url, dict(email=email))
        self._reset_cache()
        return BannedAddress(
            self._connection, response.headers.get('location'))

    def find_by_email(self, email):
        for ban in self:
            if ban.email == email:
                return ban
        return None

    def remove(self, email):
        ban = self.find_by_email(email)
        if ban is not None:
            ban.delete()
            self._reset_cache()
        else:
            raise ValueError('The address {} is not banned'.format(email))


class BannedAddress(RESTObject):

    _properties = ('email', 'list_id', 'self_link')
    _writable_properties = []

    def __repr__(self):
        return '<BannedAddress {!r}>'.format(self.email)

    def __str__(self):
        return self.email

    @property
    def mailinglist(self):
        return MailingList(
            self._connection, 'lists/{0}'.format(self.list_id))



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Harness for testing Mailman's documentation."""

from contextlib import contextmanager

__all__ = [
    'dump',
    ]


def dump(results):
    if results is None:
        print(None)
        return
    for key in sorted(results):
        if key == 'entries':
            for i, entry in enumerate(results[key]):
                # entry is a dictionary.
                print('entry %d:' % i)
                for entry_key in sorted(entry):
                    print('    {0}: {1}'.format(entry_key, entry[entry_key]))
        else:
            print('{0}: {1}'.format(key, results[key]))


@contextmanager
def print_exception(*args, **kw):
    try:
        yield
    except Exception as ex:
        print(f'{type(ex).__name__}: {ex}')



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

from mailmanclient.restbase.base import RESTObject

__metaclass__ = type
__all__ = [
    'HeldMessage'
]


class HeldMessage(RESTObject):

    _properties = ('hold_date', 'message_id', 'msg', 'reason', 'request_id',
                   'self_link', 'sender', 'subject', 'type')

    def __repr__(self):
        return '<HeldMessage {0!r} by {1}>'.format(
            self.request_id, self.sender)

    def moderate(self, action, comment=None):
        """Moderate a held message.

        :param action: Action to perform on held message.
        :type action: String.
        """
        data = dict(action=action)
        if comment is not None:
            data['comment'] = comment

        response, content = self._connection.call(
            self._url, data, 'POST')
        return response

    def discard(self):
        """Shortcut for moderate."""
        return self.moderate('discard')

    def reject(self, reason=None):
        """Shortcut for moderate.

        :param reason: An optional reason for rejecting the held message.
        """
        return self.moderate('reject', comment=reason)

    def defer(self):
        """Shortcut for moderate."""
        return self.moderate('defer')

    def accept(self):
        """Shortcut for moderate."""
        return self.moderate('accept')



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Some async method utilities."""

__all__ = [
    'list_of_objects',
]

from typing import List, TypeVar, Type
from mailmanclient.restobjects.types import ContentType, ConnectionProto
from mailmanclient.restbase.async_base import RESTObject


T = TypeVar('T', bound=RESTObject)


def list_of_objects(
        obj_type: Type[T],
        response: ContentType,
        connection: ConnectionProto) -> List[T]:
    """Convert a list of 'entries' in response to list of objects."""
    if 'entries' in response:
        entries = response.get('entries')
        if entries is not None:
            return [obj_type(connection, data) for data in entries]
    return []



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from mailmanclient.restbase.base import RESTDict

__metaclass__ = type
__all__ = [
    'Settings'
]


class Settings(RESTDict):

    _read_only_properties = (
        'bounces_address',
        'created_at',
        'digest_last_sent_at',
        'fqdn_listname',
        'join_address',
        'last_post_at',
        'leave_address',
        'list_id',
        'list_name',
        'mail_host',
        'next_digest_number',
        'no_reply_address',
        'owner_address',
        'post_id',
        'posting_address',
        'request_address',
        'scheme',
        'self_link',
        'volume',
        'web_host',
        )



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from urllib.parse import urlencode, urlsplit, parse_qs, urlunsplit

from mailmanclient.constants import DEFAULT_PAGE_ITEM_COUNT

__metaclass__ = type
__all__ = [
    'Page'
]


class Page:

    def __init__(self, connection, path, model, count=DEFAULT_PAGE_ITEM_COUNT,
                 page=1):
        self._connection = connection
        self._path = path
        self._count = count
        self._page = page
        self._model = model
        self._entries = []
        self.total_size = 0
        self._create_page()

    def __getitem__(self, key):
        return self._entries[key]

    def __iter__(self):
        for entry in self._entries:
            yield entry

    def __repr__(self):
        return '<Page {0} ({1})'.format(self._page, self._model)

    def __len__(self):
        return len(self._entries)

    def _build_url(self):
        url = list(urlsplit(self._path))
        qs = parse_qs(url[3])
        qs["count"] = self._count
        qs["page"] = self._page
        url[3] = urlencode(qs, doseq=True)
        return urlunsplit(url)

    def _create_page(self):
        self._entries = []
        response, content = self._connection.call(self._build_url())
        self.total_size = content["total_size"]
        for entry in content.get('entries', []):
            instance = self._model(
                self._connection, entry['self_link'], entry)
            self._entries.append(instance)

    @property
    def nr(self):
        return self._page

    @property
    def next(self):
        return self.__class__(
            self._connection, self._path, self._model, self._count,
            self._page + 1)

    @property
    def previous(self):
        if self.has_previous:
            return self.__class__(
                self._connection, self._path, self._model, self._count,
                self._page - 1)

    @property
    def has_previous(self):
        return self._page > 1

    @property
    def has_next(self):
        return self._count * self._page < self.total_size



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from mailmanclient.restobjects.preferences import PreferencesMixin
from mailmanclient.restbase.base import RESTObject

__metaclass__ = type
__all__ = [
    'Member'
]


class Member(RESTObject, PreferencesMixin):

    _properties = ('address', 'delivery_mode', 'email', 'list_id',
                   'moderation_action', 'display_name', 'role', 'self_link',
                   'subscription_mode', 'member_id', 'bounce_score',
                   'last_bounce_received', 'last_warning_sent',
                   'total_warnings_sent')
    _writable_properties = ('address', 'delivery_mode', 'moderation_action')

    def __repr__(self):
        return '<Member {0!r} on {1!r} with role {2!r}>'.format(
            self.email, self.list_id, self.role)

    def __str__(self):
        return 'Member "{0}" on "{1}"'.format(self.email, self.list_id)

    @property
    def address(self):
        from mailmanclient.restobjects.address import Address
        return Address(self._connection, self.rest_data['address'])

    @property
    def user(self):
        from mailmanclient.restobjects.user import User
        return User(self._connection, self.rest_data['user'])

    def unsubscribe(self):
        """Unsubscribe the member from a mailing list."""
        response, json = self._connection.call(self.self_link, method='DELETE')
        if response.status_code == 202:
            return json



# Copyright (C) 2019-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Test for the rewriting of self_link URLs."""

import unittest

from mailmanclient import Client


class TestRewriteUrl(unittest.TestCase):
    """Test that self_link paths are safely rewritten"""

    def setUp(self):
        self._client = Client(
            'http://127.0.0.1:9001/3.1', 'restadmin', 'restpass')
        self.domain = self._client.create_domain('example.org')

    def tearDown(self):
        self.domain.delete()

    def test_domain_self_link_is_localhost(self):
        domain = self._client.get_domain('example.org')
        assert domain.self_link.startswith('http://localhost:9001/')

    def test_connection_rewrite_url(self):
        link = self._client._connection.rewrite_url(self.domain.self_link)
        assert link.startswith('http://127.0.0.1:9001/')

    def test_connection_rewrites_url(self):
        response, _c = self._client._connection.call(self.domain.self_link)
        assert response.url.startswith('http://127.0.0.1:9001/')



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""User async object."""

__all__ = [
    'User',
]

from mailmanclient.restbase.async_base import RESTObject
from mailmanclient.restobjects.utils import list_of_objects
from mailmanclient.asyncobjects.preferences import PreferencesMixin


class User(RESTObject, PreferencesMixin):

    _properties = ('created_on', 'display_name', 'is_server_owner',
                   'password', 'self_link', 'user_id')
    _writable_properties = ('cleartext_password', 'display_name',
                            'is_server_owner')

    def __repr__(self):
        return '<User {0!r} ({1})>'.format(self.display_name, self.user_id)

    async def addresses(self):
        from mailmanclient.asyncobjects.address import Address
        url = self.self_link + '/addresses'
        resp, content = await self._connection.call(url)
        return list_of_objects(Address, content, self._connection)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from mailmanclient.restbase.base import RESTDict

__metaclass__ = type
__all__ = [
    'Configuration'
]


class Configuration(RESTDict):

    _writable_properties = ()

    def __init__(self, connection, name):
        super(Configuration, self).__init__(
            connection, 'system/configuration/{}'.format(name))
        self.name = name

    def __repr__(self):
        return '<Configuration: {!r}>'.format(self.name)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

from urllib.error import HTTPError

from mailmanclient.restobjects.preferences import PreferencesMixin
from mailmanclient.restobjects.address import Addresses, Address
from mailmanclient.restbase.base import RESTObject

__metaclass__ = type
__all__ = [
    'User'
]


class User(RESTObject, PreferencesMixin):

    _properties = ('created_on', 'display_name', 'is_server_owner',
                   'password', 'self_link', 'user_id')
    _writable_properties = ('cleartext_password', 'display_name',
                            'is_server_owner')

    def __init__(self, connection, url, data=None):
        super(User, self).__init__(connection, url, data)
        self._subscriptions = None
        self._subscription_list_ids = None
        self._preferred_address = None

    def __repr__(self):
        return '<User {0!r} ({1})>'.format(self.display_name, self.user_id)

    @property
    def addresses(self):
        return Addresses(
            self._connection, 'users/{0}/addresses'.format(self.user_id))

    def __setattr__(self, name, value):
        """Special case for the password"""
        if name == 'password':
            self._changed_rest_data['cleartext_password'] = value
            if self._autosave:
                self.save()
        else:
            super(User, self).__setattr__(name, value)

    @property
    def subscriptions(self):
        from mailmanclient.restobjects.member import Member
        if self._subscriptions is None:
            subscriptions = []
            for address in self.addresses:
                response, content = self._connection.call(
                    'members/find', data={'subscriber': address})
                try:
                    for entry in content['entries']:
                        subscriptions.append(Member(
                            self._connection, entry['self_link'], entry))
                except KeyError:
                    pass
            self._subscriptions = subscriptions
        return self._subscriptions

    @property
    def subscription_list_ids(self):
        if self._subscription_list_ids is None:
            list_ids = []
            for sub in self.subscriptions:
                list_ids.append(sub.list_id)
            self._subscription_list_ids = list_ids
        return self._subscription_list_ids

    def add_address(self, email, absorb_existing=False):
        """
        Adds another email adress to the user record and returns an
        _Address object.

        :param email: The address to add
        :type  email: str.
        :param absorb_existing: set this to True if you want to add the address
            even if it already exists. It will import the existing user into
            the current one, not overwriting any previously set value.
        :type  absorb_existing: bool.
        """
        url = '{0}/addresses'.format(self._url)
        data = {'email': email}
        if absorb_existing:
            data['absorb_existing'] = 1
        response, content = self._connection.call(url, data)
        address = {
            'email': email,
            'self_link': response.headers.get('location'),
        }
        return Address(self._connection, address['self_link'], address)

    @property
    def preferred_address(self):
        """Preferred address of a User.

        .. versionadded:: 3.2.3

        :returns: Address or None.
        """
        if self._preferred_address:
            return self._preferred_address
        url = '{}/preferred_address'.format(self._url)
        try:
            response, content = self._connection.call(url)
        except HTTPError as e:
            if e.code == 404:
                return None
            else:
                raise
        self._preferred_address = Address(
            self._connection, content['self_link'], content)
        return self._preferred_address

    @preferred_address.setter
    def preferred_address(self, email):
        """Set a User's preferred address.

        .. versionadded:: 3.2.3

        :param email: Email to be set as the User's preferred address. If this
            address is not linked to the user, it will be linked to the user
            unless it is linked to an other user.
            If email is None, preferred address would be unset.
        """
        url = '{}/preferred_address'.format(self._url)
        if email is None:
            response, content = self._connection.call(url, method='DELETE')
        else:
            response, content = self._connection.call(url, {'email': email})
        # Unset the preferred address we have so that we GET is next time
        # someone tries to call the attribute.
        self._preferred_address = None



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mailman.client is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import find_packages, setup

from setup_helpers import get_version, require_python

require_python(0x30600f0)
__version__ = get_version('src/mailmanclient/constants.py')


def readme():
    with open('README.rst') as fd:
        return fd.read()


setup(
    name='mailmanclient',
    version=__version__,
    packages=find_packages('src'),
    description='mailmanclient -- Python bindings for Mailman REST API',
    long_description=readme(),
    long_description_content_type='text/x-rst',
    package_dir={'': 'src'},
    include_package_data=True,
    maintainer='Barry Warsaw',
    maintainer_email='barry@list.org',
    license='LGPLv3',
    url='http://www.list.org/',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)', # noqa
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP ',
    ],
    install_requires=[
        'requests',
        'typing_extensions;python_version<"3.8"',
        ],
    extras_require={
        'testing': [
            'pytest',
            'pytest-services',
            'mailman>=3.3.1',
            'falcon>1.4.1',
            'httpx',
            ],
        'docs': [
            'sphinx',
            'sphinx-rtd-theme',
            'sphinx-issues',
            'pydoctor',
        ],
        'lint': [
            'flake8>3.0',
            'flake8-bugbear',
           ]
        },
    )



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Module contents."""


import unittest

from mailmanclient import Client


class TestTemplates(unittest.TestCase):
    """Tests for various types of templates.
    """

    def setUp(self):
        self.client = Client(
            'http://localhost:9001/3.1/',  'restadmin', 'restpass')
        self.domain = self.client.create_domain('example.net')
        self.mlist = self.domain.create_list('test1')

    def tearDown(self):
        for temp in self.mlist.templates:
            temp.delete()
        for temp in self.domain.templates:
            temp.delete()
        self.domain.delete()
        for temp in self.client.templates:
            temp.delete()

    def test_set_template(self):
        # At first, there are no templates at site, domain or list context
        self.assertEqual(len(self.client.templates), 0)
        self.assertEqual(len(self.domain.templates), 0)
        self.assertEqual(len(self.mlist.templates), 0)

        # Now, let's try to set a different template at each context level.
        self.client.set_template('list:user:notice:welcome',
                                 'http://example.com/welcome.txt')
        self.domain.set_template('list:user:notice:welcome',
                                 'http://example.com/welcome2.txt')
        self.mlist.set_template('list:user:notice:welcome',
                                'http://example.com/welcome3.txt')

        # Now, let's try to fetch the templates again.
        site_templates = self.client.templates
        self.assertEqual(len(site_templates), 1)
        self.assertEqual(site_templates[0].uri,
                         'http://example.com/welcome.txt')

        domain_templates = self.domain.templates
        self.assertEqual(len(domain_templates), 1)
        self.assertEqual(domain_templates[0].uri,
                         'http://example.com/welcome2.txt')

        mlist_templates = self.mlist.templates
        self.assertEqual(len(mlist_templates), 1)
        self.assertEqual(mlist_templates[0].uri,
                         'http://example.com/welcome3.txt')

    def test_site_templates_with_password(self):
        # It is possible to set templates with passwords.
        self.assertEqual(len(self.client.templates), 0)
        self.client.set_template('list:user:notice:goodbye',
                                 'http://example.com/goodbye.txt',
                                 username='testuser',
                                 password='testpass')
        site_templates = self.client.templates
        self.assertEqual(len(site_templates), 1)
        temp = site_templates[0]
        self.assertEqual(temp.uri, 'http://example.com/goodbye.txt')
        self.assertEqual(temp.username, 'testuser')
        self.assertEqual(temp.password, 'testpass')

    def test_domain_templates_with_password(self):
        self.assertEqual(len(self.domain.templates), 0)
        self.domain.set_template('list:user:notice:goodbye',
                                 'http://example.com/goodbye.txt',
                                 username='testuser',
                                 password='testpass')
        domain_templates = self.domain.templates
        self.assertEqual(len(domain_templates), 1)
        temp = domain_templates[0]
        self.assertEqual(temp.uri, 'http://example.com/goodbye.txt')
        self.assertEqual(temp.username, 'testuser')
        self.assertEqual(temp.password, 'testpass')

    def test_list_templates_with_password(self):
        self.assertEqual(len(self.mlist.templates), 0)
        self.mlist.set_template('list:user:notice:goodbye',
                                'http://example.com/goodbye.txt',
                                username='testuser',
                                password='testpass')
        mlist_templates = self.mlist.templates
        self.assertEqual(len(mlist_templates), 1)
        temp = mlist_templates[0]
        self.assertEqual(temp.uri, 'http://example.com/goodbye.txt')
        self.assertEqual(temp.username, 'testuser')
        self.assertEqual(temp.password, 'testpass')



# Copyright (C) 2021-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.
"""Async client for Mailman Core 3.1 API.


AsyncClient provides a thin Python API over Mailman Core's HTTP API. It is
currently is very early stages and supports on read operations. Some of the
read operations might be missinng as well.

To start using the client, you need an async http library. httpx_ is officially
supported one, but making some other client work with it is pretty easy.
::

    >>> import httpx
    >>> conn = httpx.AsyncClient()
    >>> from mailmanclient.asynclient import AsyncClient
    >>> client = AsyncClient(conn, 'http://localhost:8001/3.1',
    ...                      'restadmin', 'restpass')


You will need an event loop to actually run the client. You can reuse your
existing one by calling ``await`` on the client methods, or you can create a
new one using the standard library ``asyncio`` module.
::

    >>> import asyncio
    >>> domains = asyncio.run(client.domains())

.. _httpx: https://www.python-httpx.org/

"""

__all__ = [
    'AsyncClient',
]

from typing import List, Mapping, Any
from mailmanclient.restobjects.utils import list_of_objects
from mailmanclient.restobjects.types import HTTPClientProto
from mailmanclient.restbase.async_connection import Connection
from mailmanclient.asyncobjects.domain import Domain
from mailmanclient.asyncobjects.mailinglist import MailingList
from mailmanclient.asyncobjects.user import User
from mailmanclient.asyncobjects.address import Address
from mailmanclient.asyncobjects.member import Member


JSON_CONTENT_TYPE = 'application/json'


class AsyncClient:
    """Provide an Idiomatic API for Mailman Core.


    It requires an HTTP client instance as the first argument. You can use any
    client which has a ``.request()`` method and accepts named parameters
    ``url``, ``path``, ``auth``, ``method`` and ``data``. ``data`` is supposed
    to be a dictionary of parameters to be passed to the HTTP request and the
    rest are string parameters with their usual meaning.

    The parameters are based off on httpx python library.

    :param client: Http client object with an async request method.
    :param base_url: Base URL to Core's API.
    :param user: Core admin username.
    :param password: Core admin password.

    """

    def __init__(
        self,
        client: HTTPClientProto,
        base_url: str,
        user: str,
        password: str,
    ) -> None:
        self.client = client
        self.connection = Connection(self.client, base_url, user, password)

    async def domains(self) -> List[Domain]:
        """Get all domains.

        ``/<api>/domains``
        """
        response, content = await self.connection.call('domains')
        return list_of_objects(Domain, content, self.connection)

    async def system(self) -> Mapping[str, Any]:
        """Get the Mailman system information.

        ``/<api>/system``
        """
        response, content = await self.connection.call('system')
        return content

    async def lists(self) -> List[MailingList]:
        """Get a list of MailingLists

        ``/<api>/lists``
        """
        response, content = await self.connection.call('lists')
        return list_of_objects(MailingList, content, self.connection)

    async def members(self) -> List[Member]:
        """All the Members

        ``/<api>/members``
        """
        response, content = await self.connection.call('members')
        return list_of_objects(Member, content, self.connection)

    async def users(self) -> List[User]:
        """All the users in Mailman Core

        ``/<api>/users``
        """
        response, content = await self.connection.call('users')
        return list_of_objects(User, content, self.connection)

    async def addresses(self) -> List[Address]:
        """All the addresses in Mailman

        ``/<api>/address``
        """
        response, content = await self.connection.call('addresses')
        return list_of_objects(Address, content, self.connection)

    async def find_members(
            self, list_id: str = None, subscriber: str = None,
            role: str = None, moderation_action: str = None,
            delivery_status: str = None, delivery_mode: str = None,
    ) -> List[Member]:
        """Find members.

        ``/<api>/members/find``

        :param list_id: Mailinglist id.
        :param subscriber: Email or user_id or partial search string.
        :param role: Membership role. One of 'owner', 'member', 'nonmember' or
             'moderator'.
        :param moderation_action: One of the moderation action from 'defer',
             'accept', 'discard', 'reject', 'hold'.
        :param delivery_status: Delivery status of the Member. It can be one
             among 'enabled', 'by_user', 'by_moderator' or 'by_bounces'.
        :param delivery_mode: Delivery mode of the member. It can be one
             between 'plaintext_digests', 'mime_digests', 'regular'.
        """
        data = dict(list_id=list_id,
                    subscriber=subscriber,
                    role=role,
                    moderation_action=moderation_action,
                    delivery_status=delivery_status,
                    delivery_mode=delivery_mode)
        # Skip parameters that have None value.
        # TODO: Handle parameters that can have None value.
        data = {key: value for key, value in data.items() if value is not None}
        response, content = await self.connection.call(
            'members/find', data=data, method='GET')
        return list_of_objects(Member, content, self.connection)



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailman.client.
#
# mailman.client is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailman.client is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailman.client.  If not, see <http://www.gnu.org/licenses/>.

"""Template objects."""

from __future__ import absolute_import, print_function, unicode_literals

from mailmanclient.restbase.base import RESTList, RESTObject


__all__ = [
    'Template',
    'TemplateList'
]


class TemplateList(RESTList):

    def __init__(self, connection, url,  data=None, context=None):
        super(RESTList, self).__init__(connection, url, data)
        self._factory = lambda data: Template(
            self._connection, data['self_link'], data)


class Template(RESTObject):
    _properties = ('self_link', 'name', 'uri', 'username', 'password')
    _writable_properties = ['uri', 'username', 'password']

    def __repr__(self):
        return '<Template {}>'.format(self.name)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.
from mailmanclient.restbase.base import RESTObject

__metaclass__ = type
__all__ = [
    'Queue'
]


class Queue(RESTObject):

    _properties = ('name', 'directory', 'files')

    def __repr__(self):
        return '<Queue: {!r}>'.format(self.name)

    def inject(self, list_id, text):
        self._connection.call(self._url, dict(list_id=list_id, text=text))

    @property
    def files(self):
        # No caching.
        response, content = self._connection.call(self._url)
        return content['files']



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.


__version__ = '3.3.5'

DEFAULT_PAGE_ITEM_COUNT = 50
MISSING = object()



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

from urllib.error import HTTPError

from mailmanclient.restbase.base import RESTList, RESTObject

__metaclass__ = type
__all__ = [
    'HeaderMatch',
    'HeaderMatches'
]


class HeaderMatches(RESTList):
    """
    The list of header matches for a mailing-list.
    """

    def __init__(self, connection, url, mlist):
        """
        :param mlist: The corresponding list object.
        :type mlist: MailingList.
        """
        super(HeaderMatches, self).__init__(connection, url)
        self._mlist = mlist
        self._factory = lambda data: HeaderMatch(
            self._connection, data['self_link'], data)

    def __repr__(self):
        return '<HeaderMatches for {0!r}>'.format(self._mlist.list_id)

    def __str__(self):
        return 'Header matches for "{}"'.format(self._mlist.list_id)

    def add(self, header, pattern, action=None, tag=None):
        """Add a new HeaderMatch rule to the MailingList.

        :param header: The header to consider.
        :type  header: str
        :param pattern: The regular expression to use for filtering.
        :type  pattern: str
        :param action: The action to take when the header matches the pattern.
            This can be 'accept', 'discard', 'reject', or 'hold'.
        :type  action: str
        """
        data = dict(header=header, pattern=pattern)
        if action is not None:
            data['action'] = action
        if tag is not None:
            data['tag'] = tag
        response, content = self._connection.call(self._url, data)
        self._reset_cache()
        return HeaderMatch(self._connection, response.headers.get('location'))

    def find(self, header=None, tag=None, action=None):
        """Find a set of HeaderMatch rules.

        :param header: The header to consider.
        :type header: str
        :param tag: The tag associated with header.
        :type tag: str
        :param action: The action to take when the header matches the pattern.
            This can be 'accept', 'discard', 'reject', or 'hold'.
        :type  action: str
        """
        url = self._url + '/find'
        data = dict(header=header, tag=tag, action=action)
        data = {key: value for key, value in data.items() if value}
        if not data:
            return []
        try:
            response, content = self._connection.call(url, data)
        except HTTPError as e:
            if e.code == 404:
                return []
            raise
        return [HeaderMatch(self._connection, entry['self_link'], entry)
                for entry in content['entries']]


class HeaderMatch(RESTObject):

    _properties = ('header', 'pattern', 'position', 'action', 'tag',
                   'self_link')
    _writable_properties = ('header', 'pattern', 'position', 'action', 'tag')

    def __repr__(self):
        return '<HeaderMatch on {0!r}>'.format(self.header)

    def __str__(self):
        return 'Header match on "{}"'.format(self.header)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of mailmanclient.
#
# mailmanclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# mailmanclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with mailmanclient.  If not, see <http://www.gnu.org/licenses/>.

"""Client code."""

import warnings
from operator import itemgetter
from urllib.parse import quote

from mailmanclient.constants import (MISSING)
from mailmanclient.restobjects.address import Address
from mailmanclient.restobjects.ban import Bans, BannedAddress
from mailmanclient.restobjects.configuration import Configuration
from mailmanclient.restobjects.domain import Domain
from mailmanclient.restobjects.mailinglist import MailingList
from mailmanclient.restobjects.member import Member
from mailmanclient.restobjects.preferences import Preferences
from mailmanclient.restobjects.queue import Queue
from mailmanclient.restobjects.styles import Styles
from mailmanclient.restobjects.user import User
from mailmanclient.restobjects.templates import Template, TemplateList
from mailmanclient.restbase.connection import Connection
from mailmanclient.restbase.page import Page

__metaclass__ = type
__all__ = [
    'Client'
]


#
# --- The following classes are part of the API
#

class Client:
    """Access the Mailman REST API root.

    :param baseurl: The base url to access the Mailman 3 REST API.
    :type baseurl: str
    :param name: The Basic Auth user name.  If given, the `password` must
        also be given.
    :type name: str
    :param password: The Basic Auth password.  If given the `name` must
        also be given.
    :type password: str
    :param request_hooks: Callable hooks to process request parameters before
        being sent to Core's API.
    :type request_hooks: List[callables]
    """

    def __init__(self, baseurl, name=None, password=None, request_hooks=None):
        """Initialize client access to the REST API."""
        self._connection = Connection(baseurl, name, password, request_hooks)

    def __repr__(self):
        return '<Client ({0.name}:{0.password}) {0.baseurl}>'.format(
            self._connection)

    def add_hooks(self, request_hooks):
        """Add a hook to process connections to Mailman's API.

        Hooks are callables that are passed in the parameters of HTTP call made
        to Mailman Core' RESt API and are expected to return the same number of
        parameters back. This can be useful to manipulate parameters and update
        them, or simply log each call to API.
        ::

            from mailmanclient.client import Client
            client = Client('http://localhost:8001/3.1', <user>, <password>)

            def sample_hook(params):
                print(f'Request params are {params}')
                return params

            client.add_hook([sample_hook])


        :param request_hooks: A list of callables that take in a dictionary of
           parameters. ``params`` is the list of request parameters before the
           request is made.
        """
        self._connection.add_hooks(request_hooks=request_hooks)

    @property
    def system(self):
        """Get the basic system information.

        :returns: System information about Mailman Core
        :rtype: Dict[str, str]
        """
        return self._connection.call('system/versions')[1]

    @property
    def preferences(self):
        """Get all default system Preferences.

        :returns: System preferences.
        :rtype: :class:`Preferences`
        """
        return Preferences(self._connection, 'system/preferences')

    @property
    def configuration(self):
        """Get the system configuration.

        :returns: All the system configuration.
        :rtype: Dict[str, :class:`Configuration`]
        """
        response, content = self._connection.call('system/configuration')
        return {section: Configuration(
            self._connection, section) for section in content['sections']}

    @property
    def pipelines(self):
        """Get a list of all Pipelines.

        :returns: A list of all the pipelines in Core.
        :rtype: List
        """
        response, content = self._connection.call('system/pipelines')
        return content

    @property
    def chains(self):
        """Get a list of all the Chains.

        :returns: A list of all the chains in Core.
        :rtype: List
        """
        response, content = self._connection.call('system/chains')
        return content

    @property
    def queues(self):
        """Get a list of all Queues.

        :returns: A list of all the queues in Core.
        :rtype: List
        """
        response, content = self._connection.call('queues')
        queues = {}
        for entry in content['entries']:
            queues[entry['name']] = Queue(
                self._connection, entry['self_link'], entry)
        return queues

    @property
    def styles(self):
        """All the default styles in Mailman Core.

        :returns: All the styles in Core.
        :rtype: :class:`Styles`
        """
        return Styles(self._connection, 'lists/styles')

    @property
    def lists(self):
        """Get a list of all MailingLists.

        :returns: All the mailing lists.
        :rtype: list(:class:`MailingList`)
        """
        return self.get_lists()

    def get_lists(self, advertised=False):
        """Get a list of all the MailingLists.

        :param advertised: If marked True, returns all MailingLists including
                           the ones that aren't advertised.
        :type advertised: bool
        :returns: A list of mailing lists.
        :rtype: List(:class:`MailingList`)
        """
        url = 'lists'
        if advertised:
            url += '?advertised=true'
        response, content = self._connection.call(url)
        if 'entries' not in content:
            return []
        return [MailingList(self._connection, entry['self_link'], entry)
                for entry in content['entries']]

    def get_list_page(self, count=50, page=1, advertised=None, mail_host=None):
        """Get a list of all MailingList with pagination.

        :param count: Number of entries per-page (defaults to 50).
        :param page: The page number to return (defaults to 1).
        :param advertised: If marked True, returns all MailingLists including
                           the ones that aren't advertised.
        :param mail_host: Domain to filter results by.
        """
        if mail_host:
            url = 'domains/{0}/lists'.format(mail_host)
        else:
            url = 'lists'
        if advertised:
            url += '?advertised=true'
        return Page(self._connection, url, MailingList, count, page)

    @property
    def domains(self):
        """Get a list of all Domains.

        :returns: All the domains on the system.
        :rtype: List[:class:`Domain`]
        """
        response, content = self._connection.call('domains')
        if 'entries' not in content:
            return []
        return [Domain(self._connection, entry['self_link'])
                for entry in sorted(content['entries'],
                                    key=itemgetter('mail_host'))]

    @property
    def members(self):
        """Get a list of all the Members.

        :returns: All the list memebrs.
        :rtype: List[:class:`Member`]
        """
        response, content = self._connection.call('members')
        if 'entries' not in content:
            return []
        return [Member(self._connection, entry['self_link'], entry)
                for entry in content['entries']]

    def get_member(self, fqdn_listname, subscriber_address):
        """Get the Member object for a given MailingList and Subsciber's Email
        Address.

        :param str fqdn_listname: Fully qualified address for the MailingList.
        :param str subscriber_address: Email Address for the subscriber.
        :returns: A member of a list.
        :rtype: :class:`Member`
        """
        return self.get_list(fqdn_listname).get_member(subscriber_address)

    def get_nonmember(self, fqdn_listname, nonmember_address):
        """Get the Member object for a given MailingList and Non-member's
        Email.

        :param str fqdn_listname: Fully qualified address for the MailingList.
        :param str subscriber_address: Email Address for the non-member.
        :returns: A member of a list.
        :rtype: :class:`Member`
        """
        return self.get_list(fqdn_listname).get_nonmember(nonmember_address)

    def get_member_page(self, count=50, page=1):
        """Return a paginated list of Members.

        :param int count: Number of items to return.
        :param int page: The page number.
        :returns: Paginated lists of members.
        :rtype: :class:`Page` of :class:`Member`.
        """
        return Page(self._connection, 'members', Member, count, page)

    @property
    def users(self):
        """Get all the users.

        :returns: All the users in Mailman Core.
        :rtype: List[:class:`User`]
        """
        response, content = self._connection.call('users')
        if 'entries' not in content:
            return []
        return [User(self._connection, entry['self_link'], entry)
                for entry in sorted(content['entries'],
                                    key=itemgetter('self_link'))]

    def get_user_page(self, count=50, page=1):
        """Get all the users with pagination.

        :param int count: Number of entries per-page (defaults to 50).
        :param int page: The page number to return (defaults to 1).
        :returns: Paginated list of users on Mailman.
        :rtype: :class:`Page` of :class:`User`
        """
        return Page(self._connection, 'users', User, count, page)

    def create_domain(self, mail_host, base_url=MISSING,
                      description=None, owner=None, alias_domain=None):
        """Create a new Domain.

        :param str mail_host: The Mail host for the new domain. If you want
            foo@bar.com" as the address for your MailingList, use "bar.com"
            here.
        :param str description: A brief description for this Domain.
        :param str owner: Email address for the owner of this list.
        :param str alias_domain: Alias domain.
        :returns: The created Domain.
        :rtype: :class:`Domain`
        """
        if base_url is not MISSING:
            warnings.warn(
                'The `base_url` parameter in the `create_domain()` method is '
                'deprecated. It is not used any more and will be removed in '
                'the future.', DeprecationWarning, stacklevel=2)
        data = dict(mail_host=mail_host)
        if description is not None:
            data['description'] = description
        if owner is not None:
            data['owner'] = owner
        if alias_domain is not None:
            data['alias_domain'] = alias_domain
        response, content = self._connection.call('domains', data)
        return Domain(self._connection, response.headers.get('location'))

    def delete_domain(self, mail_host):
        """Delete a Domain.

        :param str mail_host: The Mail host for the domain you want to delete.
        """
        response, content = self._connection.call(
            'domains/{0}'.format(mail_host), None, 'DELETE')

    def get_domain(self, mail_host, web_host=MISSING):
        """Get Domain by its mail_host."""
        if web_host is not MISSING:
            warnings.warn(
                'The `web_host` parameter in the `get_domain()` method is '
                'deprecated. It is not used any more and will be removed in '
                'the future.', DeprecationWarning, stacklevel=2)
        response, content = self._connection.call(
            'domains/{0}'.format(mail_host))
        return Domain(self._connection, content['self_link'])

    def create_user(self, email, password, display_name=''):
        """Create a new User.

        :param str email: Email address for the new user.
        :param str password: Password for the new user.
        :param str display_name: An optional name for the new user.
        :returns: The created user instance.
        :rtype: :class:`User`
        """
        response, content = self._connection.call(
            'users', dict(email=email,
                          password=password,
                          display_name=display_name))
        return User(self._connection, response.headers.get('location'))

    def get_user(self, address):
        """Given an Email Address, return the User it belongs to.

        :param str address: Email Address of the User.
        :returns: The user instance that owns the address.
        :rtype: :class:`User`
        """
        response, content = self._connection.call(
            'users/{0}'.format(address))
        return User(self._connection, content['self_link'], content)

    def get_address(self, address):
        """Given an Email Address, return the Address object.

        :param str address: Email address.
        :returns: The Address object for given email address.
        :rtype: :class:`Address`
        """
        response, content = self._connection.call(
            'addresses/{0}'.format(address))
        return Address(self._connection, content['self_link'], content)

    def get_list(self, fqdn_listname):
        """Get a MailingList object.

        :param str fqdn_listname: Fully qualified name of the MailingList.
        :returns: The mailing list object of the given fqdn_listname.
        :rtype: :class:`MailingList`
        """
        response, content = self._connection.call(
            'lists/{0}'.format(fqdn_listname))
        return MailingList(self._connection, content['self_link'], content)

    def delete_list(self, fqdn_listname):
        """Delete a MailingList.

        :param str fqdn_listname: Fully qualified name of the MailingList.
        """
        response, content = self._connection.call(
            'lists/{0}'.format(fqdn_listname), None, 'DELETE')

    @property
    def bans(self):
        """Get a list of all the bans.

        :returns: A list of all the bans.
        :rtype: :class:`Bans`
        """
        return Bans(self._connection, 'bans', mlist=None)

    def get_bans_page(self, count=50, page=1):
        """Get a list of all the bans with pagination.

        :param int count: Number of entries per-page (defaults to 50).
        :param int page: The page number to return (defaults to 1).
        :returns: Paginated list of banned addresses.
        :rtype: :class:`Page` of :class:`BannedAddress`
        """
        return Page(self._connection, 'bans', BannedAddress, count, page)

    @property
    def templates(self):
        """Get all site-context templates.

        :returns: List of templates for the site context.
        :rtype: :class:`TemplateList`
        """
        return TemplateList(self._connection, 'uris')

    def get_templates_page(self, count=25, page=1):
        """Get paginated site-context templates.

        :returns: Paginated list of templates of site context.
        :rtype: :class:`Page` of :class:`Template`
        """
        return Page(self._connection, 'uris', Template, count, page)

    def set_template(self, template_name, url, username=None, password=None):
        """Set template in site-context.

        :param str template_name: The template to set.
        :param str url: The URL to fetch the template from.
        :param str username: Username for access to the template.
        :param str password: Password for the ``username`` to access templates.
        """
        data = {template_name: url}
        if username is not None and password is not None:
            data['username'] = username
            data['password'] = password
        return self._connection.call('uris', data, 'PATCH')[1]

    def find_lists(self, subscriber, role=None, count=50, page=1,
                   mail_host=None):
        """Given a subscriber and a role, return all the list they are
        subscribed to with given role.

        If no role is specified all the related mailing lists are returned
        without duplicates, even though there can potentially be multiple
        memberships of a user in a single mailing list.

        :param str subscriber: The address of the subscriber.
        :param str role: owner, moderator or subscriber.
        :param int count: Number of entries per-page (defaults to 50).
        :param int page: The page number to return (defaults to 1).
        :param str mail_host: Domain to filter results by.
        :returns: A filtered list of mailing lists with given filters.
        :rtype: List[:class:`MailingList`]
        """
        url = 'lists/find'
        data = dict(subscriber=subscriber, count=count, page=page)
        if role is not None:
            data['role'] = role
        response, content = self._connection.call(url, data)
        if 'entries' not in content:
            return []
        return [MailingList(self._connection, entry['self_link'], entry)
                for entry in content['entries']
                if not mail_host or entry['mail_host'] == mail_host]

    def find_users(self, query):
        """Find users with query string matching display name and emails.

        :param str query: The string to search for. The search string is case
            insensitive as core doens't care about the case.
        :param int count: Number of entries per-page (defaults to 50).
        :param int page: The page number to return (defaults to 1).
        """
        url = 'users/find?q={}'.format(quote(query))
        response, content = self._connection.call(url)
        if 'entries' not in content:
            return []
        return [User(self._connection, entry['self_link'], entry)
                for entry in content['entries']]

    def find_users_page(self, query, count, page):
        """Same as :py:meth:`find_users` but allows for pagination.

        :param str query: The string to search for. The search string is case
            insensitive as core doens't care about the case.
        :param int count: Number of entries per-page (defaults to 50).
        :param int page: The page number to return (defaults to 1).
        """
        url = 'users/find?q={}'.format(quote(query))
        return Page(self._connection, url, User, count, page)
